--[[****************************************************************
	LootCount DropCount v0.54

	Author: Evil Duck
	****************************************************************

	For the game World of Warcraft
	Stand-alone addon, and plug-in for the add-on LootCount.

	****************************************************************]]

-- 0.54 Updates the done quests with server-provided data, bugfix in
--      profession-loot, added mob-update, show/hide quests in maps
-- 0.50 Coloured quest-titles, tweaked cache speed, upgraded to DuckMod 2.0,
--      added support for profession-loot from mobs, added prebuilt database
-- 0.40 Added quests-givers, tracking of quests done, where quests start
-- 0.32 Added repair-vendor display
-- 0.30b Quick bugfix
-- 0.30a Fixed dependency-bug
-- 0.30 Added vendors: minimap/worldmap/search, added books: minimap/worldmap/search
-- 0.20 Added support for sharing data in guild, some bugfixes
-- 0.10 First version


LOOTCOUNT_DROPCOUNT_VERSIONTEXT = "DropCount v0.54";
LOOTCOUNT_DROPCOUNT = "DropCount";
SLASH_DROPCOUNT1 = "/dropcount";
SLASH_DROPCOUNT2 = "/lcdc";
local DuckLib=DuckMod[2.0];
local LootCount_DropCount_Debug=nil;
local LootCount_DropCount_Registered=nil;
local LootCount_DropCount_Target=nil;
local LootCount_DropCount_GUID="0x0";
local LootCount_DropCount_SkinTarget=nil;
local LootCount_DropCount_SkinTargetKill=nil;
local LootCount_DropCount_LootedMobs={};
local LootCount_DropCount_SkinnedMobs={};
local LootCount_DropCount_ReceivedSkins={
	A="none",
	B="none",
	C="none",
};
local LCDC_LOOTEDAGE=1200;
local LCDC_PERMANENTITEM=-1;
local LCDC_UNKNOWNCOUNT=-2;
local LCDC_MaintainerSPV=nil;
local LCDC_MaintainerSPI=nil;
local LCDC_MaintainerPOS=0;
local LCDC_LastUpdate=0;
-- Saved per character
LootCount_DropCount_Character={
	Skinning=nil,
	ShowZone=true,
};
-- Global save
LootCount_DropCount_DB={
	Item={},
	Count={},
	Vendor={},
	Book={},
	Quest={},
};
LootCount_DropCount_NoQuest = {
	[2735] = true,
	[10593] = true,
	[2751] = true,
	[8705] = true,
	[21377] = true,
	[8391] = true,
	[2799] = true,
	[8392] = true,
	[16656] = true,
	[8393] = true,
	[11512] = true,
	[8394] = true,
	[8396] = true,
	[5113] = true,
	[12840] = true,
	[22528] = true,
	[22529] = true,
	[28452] = true,
	[29209] = true,
	[2744] = true,
	[8483] = true,
	[22526] = true,
	[18944] = true,
	[21383] = true,
	[11018] = true,
	[22527] = true,
	[12841] = true,
	[10450] = true,
	[2732] = true,
	[5117] = true,
	[35188] = true,
	[4582] = true,
	[2740] = true,
	[2748] = true,
	[5134] = true,
	[11407] = true,
	[29425] = true,
	[30810] = true,
	[2725] = true,
	[31812] = true,
	[2738] = true,
	[2749] = true,
	[38551] = true,
	[30809] = true,
	[29739] = true,
	[29740] = true,
	[29426] = true,
	[2734] = true,
	[2742] = true,
	[2750] = true,
	[24291] = true,
	[24401] = true,
	[25433] = true,
	[2730] = true,
	[25719] = true,
}

-- AARRGGBB
local LCDC_C_BASIC="|cFF00FFFF";
local LCDC_C_GREEN="|cFF00FF00";
local LCDC_C_RED="|cFFFF0000";
local LCDC_C_LBLUE="|cFF6060FF";
local LCDC_C_HBLUE="|cFEA0A0FF";		-- highlight
local LCDC_C_YELLOW="|cFFFFFF00";

local LCDC_LISTLENGTH=25;
local LCDC_KILLS_HIDE=10;
local LCDC_KILLS_SAFE=50;

local LCDC_RESCANQUESTS=3;

local LCDC_UnitType=nil;
local LootCount_DropCount_LootList={};
local LootCount_DropCount_TimedQueue = {};
local LootCount_DropCount_QueueSize=0;
local LCDC_QUEUETIME=900;
local LCT_QUEST=nil;
local LootCount_DropCount_EnumerateTimer=6;
local LootCount_DropCount_CacheRetries=0;
local LCDC_ENUMERATE=1/3;
local LCSV_MAXRETRIES=10;
local LCR_QUEST=-1;
local LootCount_DropCount_This=nil;
local LCDC_PREFIX="LcDc";
local LCDC_SEPARATOR=",";
local LCDC_MOBKILL="MOB_KILL";
local LCDC_MOBLOOT="MOB_LOOT";
local LCDC_CachedConvertItems=0;
local LCDC_QuestConvertItems=0;
local LootCount_DropCount_QueryItem=nil;
local LootCount_DropCount_VendorDelay=-1;
local LootCount_DropCount_VendorProblem=nil;
local LootCount_DropCount_UnknownItems=nil;
local LootCount_DropCount_RunningItem=nil;
local LootCount_DropCount_LastID=0;
local LootCount_DropCount_LastLootID=nil;
local LootCount_DropCount_LastAliveTarget=nil;
local LootCount_DropCount_CurrentAliveFriendClose=nil;
local LootCount_DropCount_OpenMerchant=nil;
local LootCount_DropCount_Incoming={};
local LootCount_DropCount_Profession=nil;
local LootCount_DropCount_Professions={};
local LootCount_DropCount_SkinningList={
	[2934]=true,		-- Ruined Leather Scraps
	[2318]=true,		-- Light Leather
	[2319]=true,		-- Medium Leather
	[4234]=true,		-- Heavy Leather
	[4304]=true,		-- Thick Leather
	[8170]=true,		-- Rugged Leather
	[25649]=true,		-- Knothide Leather Scraps
	[21887]=true,		-- Knothide Leather
	[5082]=true,		-- Thin Kodo Leather
	[15423]=true,		-- Chimera Leather
	[15417]=true,		-- Devilsaur Leather
	[15422]=true,		-- Frostsaber Leather
	[15419]=true,		-- Warbear Leather
	[25703]=true,		-- Zhevra Leather
	[19767]=true,		-- Primal Bat Leather
	[19768]=true,		-- Primal Tiger Leather
	[17012]=true,		-- Core Leather
	[25699]=true,		-- Crystal Infused Leather
	[25708]=true,		-- Thick Clefthoof Leather
	[25700]=true,		-- Fel Scales
	[29539]=true,		-- Cobra Scales
	[29548]=true,		-- Nether Dragonscales
	[783]=true,			-- Light Hide
	[4232]=true,		-- Medium Hide
	[4235]=true,		-- Heavy Hide
	[8169]=true,		-- Thick Hide
	[8171]=true,		-- Rugged Hide
	[25707]=true,		-- Fel Hide
	[7428]=true,		-- Shadowcat Hide
	[8368]=true,		-- Thick Wolfhide
	[12731]=true,		-- Pristine Hide of the Beast
	[20500]=true,		-- Light Silithid Carapace
	[20501]=true,		-- Heavy Silithid Carapace
	[8167]=true,		-- Turtle Scale
	[8154]=true,		-- Scorpid Scale
	[15408]=true,		-- Heavy Scorpid Scale
	[20498]=true,		-- Silithid Chitin
	[29547]=true,		-- Wind Scales
	[7072]=true,		-- Naga Scale
	[7286]=true,		-- Black Whelp Scale
	[7287]=true,		-- Red Whelp Scale
	[7392]=true,		-- Green Whelp Scale
	[12607]=true,		-- Brilliant Chromatic Scale
	[8165]=true,		-- Worn Dragonscale
	[15412]=true,		-- Green Dragonscale
	[15415]=true,		-- Blue Dragonscale
	[15416]=true,		-- Black Dragonscale
	[15414]=true,		-- Red Dragonscale
	[15410]=true,		-- Scale of Onyxia
	[20381]=true,		-- Dreamscale
	[6470]=true,		-- Deviate Scale (drop and skinning)
	[6471]=true,		-- Perfect Deviate Scale (drop and skinning)
	[33567]=true,		-- Borean Leather Scraps
	[33568]=true,		-- Borean Leather
--	[12607]=true,		-- Brilliant Chromatic Scale
	[44128]=true,		-- Arctic Fur
};

LCDC_VendorIcon={};
LCDC_VendorIconMM={};
LCDC_BookIcon={};
LCDC_BookIconMM={};
LCDC_QuestIcon={};
LCDC_QuestIconMM={};
local LCDC_ItemTextFrameHandled=nil;
local LCDC_Zones=nil;
local LCDC_StartupDelay=5;
local LCDC_Minimap_Menu=nil;
local LCDC_MyFaction=nil;
local LCDC_LastFaction=nil;
local LCDCQI_STATUS_UNKNOWN=0;
local LCDCQI_STATUS_DONE=1;
local LCDCQI_STATUS_STARTED=2;
local LCDCQI_STATUS_NOTSTARTED=3;
local QuestsDone = {
	Waiter=-1,
};

-- reference to the Astrolabe mapping library
local Astrolabe = DongleStub("Astrolabe-0.4");

-- Set up for handling
function LootCount_DropCount_OnLoad()
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	this:RegisterEvent("PLAYER_FOCUS_CHANGED");
	this:RegisterEvent("PLAYER_TARGET_CHANGED");
	this:RegisterEvent("CHAT_MSG_ADDON");
	this:RegisterEvent("CHAT_MSG_CHANNEL");
	this:RegisterEvent("LOOT_OPENED");
	this:RegisterEvent("LOOT_SLOT_CLEARED");
	this:RegisterEvent("MERCHANT_SHOW");
	this:RegisterEvent("MERCHANT_CLOSED");
	this:RegisterEvent("WORLD_MAP_UPDATE");
	this:RegisterEvent("ZONE_CHANGED_NEW_AREA");
	this:RegisterEvent("QUEST_DETAIL");
	this:RegisterEvent("QUEST_COMPLETE");
	this:RegisterEvent("QUEST_FINISHED");
	this:RegisterEvent("UNIT_SPELLCAST_START");
	this:RegisterEvent("QUEST_QUERY_COMPLETE");

	local name;
	name=GetSpellInfo(8613); LootCount_DropCount_Professions[1]=name;	-- Skinning
--DuckLib:Chat("Skinning: "..name,1,0,1);
	name=GetSpellInfo(2366); LootCount_DropCount_Professions[2]=name;	-- Herb gathering
--DuckLib:Chat("Herb gathering: "..name,1,0,1);
	name=GetSpellInfo(2575); LootCount_DropCount_Professions[3]=name;	-- Mining
--DuckLib:Chat("Mining: "..name,1,0,1);
	name=GetSpellInfo(49383); LootCount_DropCount_Professions[4]=name;	-- Salvaging
--DuckLib:Chat("Salvaging: "..name,1,0,1);

	StaticPopupDialogs["LCDC_D_NOTIFICATION"] = {
		text="Text",
		button1="Close",
		OnAccept = function()
			StaticPopup_Hide ("LCDC_D_NOTIFICATION");
		end,
		timeout=0,
		whileDead=1,
		hideOnEscape=1,
	};

	SlashCmdList["DROPCOUNT"]=function(msg) LootCount_DropCount_Slasher(msg) end;

	LCDC_Minimap_Menu=CreateFrame("Frame","LCDC_Minimap_MenuFrame",Minimap,"UIDropDownMenuTemplate");
	LCDC_Minimap_Menu:SetPoint("LEFT",Minimap);
	UIDropDownMenu_Initialize(LCDC_Minimap_Menu,LCDC_Minimap_Menu_Initialise,"MENU");

	DuckLib:Init();
end


-- There's slashing to be done
function LootCount_DropCount_Slasher(msg)
	if (not msg) then msg=""; end
	if (strlen(msg)>0) then msg=strlower(msg); end

	if (msg=="guild") then
		LootCount_DropCount_DB.GUILD=true;
		LootCount_DropCount_DB.RAID=nil;
		DuckLib:Chat("DropCount: Sharing data with guild");
		return;
	elseif (msg=="raid" or msg=="group") then
		LootCount_DropCount_DB.GUILD=nil;
		LootCount_DropCount_DB.RAID=true;
		DuckLib:Chat("DropCount: Sharing data with party/raid");
		return;
	elseif (msg=="noshare") then
		LootCount_DropCount_DB.GUILD=nil;
		LootCount_DropCount_DB.RAID=nil;
		DuckLib:Chat("DropCount: Data-sharing OFF");
		return;
	elseif (msg=="zone") then
		LootCount_DropCount_Menu_ToggleZoneMobs()
		if (LootCount_DropCount_Character.ShowZoneMobs) then
			DuckLib:Chat("DropCount: Showing drop-data from current zone only");
		else
			DuckLib:Chat("DropCount: Showing drop-data from all zones");
		end
		return;
	elseif (string.find(msg,"item")==1) then
		msg=string.sub(msg,5);
		msg=strtrim(msg);
		LootCount_DropCount_QueryItem=msg;
		local number=tonumber(LootCount_DropCount_QueryItem);
		number=tostring(number);
		if (LootCount_DropCount_QueryItem==number) then
			LootCount_DropCount_QueryItem="item:"..LootCount_DropCount_QueryItem..":0:0:0:0:0:0";
		end
		if (GetItemInfo(LootCount_DropCount_QueryItem)) then			-- Fastest check
			LootCount_DropCount_ItemVendors();
		else
			if (not string.find(LootCount_DropCount_QueryItem,"item:")) then	-- Not ID, not link, assume name
				LootCount_DropCount_QueryItem=LootCount_DropCount_FindByName(LootCount_DropCount_QueryItem);
				if (not LootCount_DropCount_QueryItem) then				-- Can't find by name
					DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_RED.."Information not available for this item. A link or itemID is required.");
					LootCount_DropCount_QueryItem=nil;
					return;
				end
			end
			DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_LBLUE.."Getting item information...");
			LootCount_DropCount_StartCache(LootCount_DropCount_QueryItem);
		end
		return;
	elseif (string.find(msg,"book")==1) then
		msg=string.sub(msg,5);
		msg=strtrim(msg);
		if (msg=="zone") then
			LootCount_DropCount_ListZoneBooks();
		else
			LootCount_DropCount_ListBook(msg);
		end
		return;
	elseif (msg=="show") then
		LCDC_VendorSearch:Show();
		return;
	end

	if (msg=="debug") then
		if (LootCount_DropCount_Debug) then
			LootCount_DropCount_Debug=nil;
			DuckLib:Chat("DropCount: Debug: OFF");
		else
			LootCount_DropCount_Debug=true;
			DuckLib:Chat("DropCount: Debug: ON");
		end
		return;
	end

	DuckLib:Chat(LCDC_C_BASIC..LOOTCOUNT_DROPCOUNT_VERSIONTEXT.."|r");
	if (msg=="?") then
		LootCount_DropCount_CleanDB();
		if (LootCount_DropCount_DB.GUILD) then DuckLib:Chat(LCDC_C_BASIC.."DC:|r Currently sharing data with "..LCDC_C_GREEN.."guild|r");
		elseif (LootCount_DropCount_DB.RAID) then DuckLib:Chat(LCDC_C_BASIC.."DC:|r Currently sharing data with "..LCDC_C_LBLUE.."party/raid|r");
		else DuckLib:Chat(LCDC_C_BASIC.."DC:|r Currently "..LCDC_C_RED.."not|r sharing data"); end
		return;
	end

	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." ?|r -> Statistics");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." guild|r -> Share data with guild");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." raid|r -> Share data with party/raid");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." noshare|r -> Do not share data");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." item <item-ID\|link\|name>|r -> List vendors with this item");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." book <title>|r -> List location(s) of this book");
	DuckLib:Chat(LCDC_C_GREEN..SLASH_DROPCOUNT2.." book zone -> List all known books in current zone");
end


function LootCount_DropCount_SetVendorIntegrity()
end


-- Structure after merge:
--		Vendor
--			Zone = Greater zone name (Zangarmarsh, etc)
--			Faction = faction
--			Items
--				item:XXXX:0:0:0:0:0:0
--					Count = -2
--					Name = <missing if not cached>
--			X = 0
--			Y = 0


function LootCount_DropCount_Maintainer(start)
	if (not LootCount_DropCount_DB.Converted or (LCDC_MaintainerPOS<=0 and not start)) then return; end

	if (LCDC_MaintainerPOS==1) then
		local venAll,venNew=0,0;
		if (LCDC_MaintainerSPV and LCDC_MaintainerSPI) then venAll,venNew=LootCount_DropCount_Maintainer_SP(); end
		if (venNew>0) then DuckLib:Chat("Database maintenance: "..venAll.."/"..venNew,0,1,0); end
	elseif (LCDC_MaintainerPOS==100) then
--		if (LootCount_DropCount_UnknownItems) then LCDC_MaintainerPOS=LCDC_MaintainerPOS+1; return; end
--[[
		local missing,queued=0,0;
		if (LootCount_DropCount_DB.Vendor) then
			for itsname,vt in pairs(LootCount_DropCount_DB.Vendor) do
				if (vt.Items) then
					for i,it in pairs(vt.Items) do
						if (not it.Name) then
							missing=missing+1;
							local itemname,itemlink=GetItemInfo(i);
							if (not itemname or not itemlink) then
								LootCount_DropCount_StartCache(i);
								queued=queued+1;
							else
								vt.Items[i].Name=itemname;
							end
						end
					end
				end
			end
		end
		DuckLib:Chat("Database maintenance: m"..missing.." q"..queued,0,1,0);
]]
--		if (missing>0) then return; end
	elseif (LCDC_MaintainerPOS==101) then
--		if (LootCount_DropCount_UnknownItems) then return; end
	elseif (LCDC_MaintainerPOS>101) then
--		DuckLib:Chat("Database integrity: OK",0,1,0);
		LCDC_MaintainerPOS=-1; return;
	end
	LCDC_MaintainerPOS=LCDC_MaintainerPOS+1;
end


function LootCount_DropCount_Maintainer_SP()
	if (not LootCount_DropCount_DB.Vendor) then LootCount_DropCount_DB.Vendor={}; end
	local venAll,venNew=0,0;
	-- Spool vendors
	for name,place in pairs(LCDC_MaintainerSPV) do
		venAll=venAll+1;
		if (not LootCount_DropCount_DB.Vendor[name]) then
			LootCount_DropCount_DB.Vendor[name]={ Zone=place, X=0, Y=0 };
			venNew=venNew+1;
		end
	end
	-- Stuff items and faction
	for faction,fTable in pairs(LCDC_MaintainerSPI) do
		for item,iTable in pairs(fTable) do
			for _,vendor in pairs(iTable) do
				local here=LootCount_DropCount_DB.Vendor[vendor];
				if (not here.Faction) then here.Faction=faction; end
				if (not here.Items) then here.Items={}; end
				if (not here.Items["item:"..item..":0:0:0:0:0:0"]) then
					here.Items["item:"..item..":0:0:0:0:0:0"]={ Count=-2 };
				end
			end
		end
	end
	return venAll,venNew;
end


-- An event has been received
function LootCount_DropCount_OnEvent(event)
	if (event=="COMBAT_LOG_EVENT_UNFILTERED") then
		if (arg2=="PARTY_KILL" and (bit.band(arg3,COMBATLOG_OBJECT_TYPE_PET) or bit.band(arg3,COMBATLOG_OBJECT_TYPE_PLAYER))) then
			if (GetNumPartyMembers()<1) then
				LootCount_DropCount_AddKill(true,arg6,arg7,LootCount_DropCount_Character.Skinning);
			end
			if (LootCount_DropCount_Registered) then LootCountAPI.Force(LOOTCOUNT_DROPCOUNT); end
		end
		return;
	end

	if (event=="PLAYER_FOCUS_CHANGED" or event=="PLAYER_TARGET_CHANGED") then
		local targettype=LootCount_DropCount_GetTargetType();
		if (not targettype) then
			LootCount_DropCount_Target=nil;
			LootCount_DropCount_SkinTarget=nil;
			LootCount_DropCount_SkinTargetKill=nil;
			LootCount_DropCount_CurrentAliveFriendClose=nil;
			return;
		end
		LootCount_DropCount_Target=nil;
		LootCount_DropCount_LastLootID=nil;
		if (not UnitIsDead(targettype)) then
			LCDC_LastFaction=UnitFactionGroup(targettype);
			if (not LCDC_LastFaction) then LCDC_LastFaction="Neutral"; end
			LootCount_DropCount_LastAliveTarget=UnitName(targettype);
			LootCount_DropCount_CurrentAliveFriendClose=nil;
			if (CheckInteractDistance(targettype,2)) then	-- Trade-distance
				LootCount_DropCount_CurrentAliveFriendClose=LootCount_DropCount_LastAliveTarget;
			end
			return;
		end
		if (UnitIsFriend("player",targettype)) then return; end
		LootCount_DropCount_CurrentAliveFriendClose=nil;
		LCDC_UnitType=LCDC_GetUnitType(targettype);
		LootCount_DropCount_GUID=UnitGUID(targettype);
		LootCount_DropCount_SkinTarget=UnitName(targettype);				-- Get current valid target
		LootCount_DropCount_SkinTargetKill=LootCount_DropCount_SkinTarget;	-- Set unit for skinning-drop
		if (UnitIsTapped(targettype)) and (not UnitIsTappedByPlayer(targettype)) then return; end	-- Not my kill (in case of skinning)
		LootCount_DropCount_Target=LootCount_DropCount_SkinTarget;			-- Save name of dead targetted/focused enemy
		LootCount_DropCount_LastLootID=time();
		return;
	end

	if (event=="LOOT_OPENED") then
		LootCount_DropCount_LootList=nil; LootCount_DropCount_LootList={};
		local slots=GetNumLootItems();
		if (slots<1) then return; end

--if (LootCount_DropCount_Profession) then
--	DuckLib:Chat("This loot comes from \""..LootCount_DropCount_Profession.."\"",1,0,1);
--else
--	DuckLib:Chat("This is normal loot",1,0,1);
--end

		for i=1,slots do
			LootCount_DropCount_LootList[i]={};
			_,_,LootCount_DropCount_LootList[i].Count=GetLootSlotInfo(i);	-- Returns icon path, item name, and item quantity for the item in the given loot window slot
			LootCount_DropCount_LootList[i].Item=DuckLib:GetID(GetLootSlotLink(i));	-- Returns an itemLink for the item in the given loot window slot
		end
		local mTable=LootCount_DropCount_LootedMobs;	-- Set normal loot mobs
		if (LootCount_DropCount_Profession) then
			mTable=LootCount_DropCount_SkinnedMobs;		-- Set skinning loot mobs
		else
			-- It's normal, so check if it has already been skinned
			if (LootCount_DropCount_SkinnedMobs[LootCount_DropCount_GUID]) then return; end			-- Loot already done for this one
		end
		if (mTable[LootCount_DropCount_GUID]) then return; end			-- Loot already done for this one
		if (LootCount_DropCount_Profession and LootCount_DropCount_Target) then		-- If my kill
			LootCount_DropCount_AddKill(true,LootCount_DropCount_GUID,LootCount_DropCount_Target);
		end

		local now=time();
		-- Save loot
		mTable[LootCount_DropCount_GUID]=now;							-- Set it
		for i=1,slots do
			if (LootCount_DropCount_LootList[i].Count>0) then			-- Not money
				if (LootCount_DropCount_Target) then
					LootCount_DropCount_AddLocalLoot(LootCount_DropCount_GUID,LootCount_DropCount_Target,LootCount_DropCount_LootList[i].Item,LootCount_DropCount_LootList[i].Count);
				elseif (LootCount_DropCount_SkinTarget) then
					LootCount_DropCount_AddLocalLoot(LootCount_DropCount_GUID,LootCount_DropCount_SkinTarget,LootCount_DropCount_LootList[i].Item,LootCount_DropCount_LootList[i].Count);
				end
			end
		end
		LootCount_DropCount_Profession=nil;			-- Set normal type loot
		-- Remove old mobs
		for guid,when in pairs(mTable) do
			if (now-when>LCDC_LOOTEDAGE) then mTable[guid]=nil; end
		end
		return;
	end

	if (event=="CHAT_MSG_ADDON") then
		if (arg1~=LCDC_PREFIX) then return; end
		if (LootCount_DropCount_DB.RAID) then
			if (arg3~="RAID" and arg3~="PARTY") then return; end
		elseif (LootCount_DropCount_DB.GUILD) then
			if (arg3~="GUILD") then return; end
		else return; end

		if (arg4==UnitName("player")) then return; end
		LootCount_DropCount_ParseMessage(arg2,arg4);
	end

	if (event=="MERCHANT_SHOW") then
		LootCount_DropCount_OpenMerchant=LootCount_DropCount_LastAliveTarget;
		LootCount_DropCount_VendorDelay=.5;
		return;
	end
	if (event=="MERCHANT_CLOSED") then
		LootCount_DropCount_VendorDelay=-1;
		LootCount_DropCount_VendorProblem=nil;
		return;
	end

	if (event=="WORLD_MAP_UPDATE") then LootCount_DropCount_PlotIcons(); return; end
	if (event=="ZONE_CHANGED_NEW_AREA") then
		LootCount_DropCount_MakeIcons();
		LootCount_DropCount_MakeIconsBook();
		LootCount_DropCount_MakeIconsQuest();
		return;
	end

	if (event=="QUEST_DETAIL") then
		local qName=GetTitleText();
		local target=LootCount_DropCount_GetTargetType();
		if (target) then
			target=CheckInteractDistance(LootCount_DropCount_GetTargetType(),2);	-- Trade-distance
		end
		if (not target) then		-- No target, or too far away
			LootCount_DropCount_CurrentAliveFriendClose=nil;
		end
		if (qName) then LCDC_SaveQuest(qName,LootCount_DropCount_CurrentAliveFriendClose); end

--[[
		local qList={ GetGossipAvailableQuests() };
--name, level, isTrivial, ... =
		local i=1;
		while (qList[i]) do
			LCDC_SaveQuest(qList[i],LootCount_DropCount_LastAliveTarget,qList[i+1]);
			i=i+3;
		end
		DuckLib:Chat(i,1);
]]
		return;
	end

	if (event=="QUEST_COMPLETE") then
		local qName=GetTitleText();
		local qID=true;
		if (LootCount_DropCount_Character.Quests[qName]) then
			qID=LootCount_DropCount_Character.Quests[qName];
		end
		if (not LootCount_DropCount_Character.DoneQuest) then LootCount_DropCount_Character.DoneQuest={}; end
		if (not LootCount_DropCount_Character.DoneQuest[qName]) then
			DuckLib:Chat("Quest \""..qName.."\" completed",0,1,0);
			LootCount_DropCount_Character.DoneQuest[qName]=qID;
		else
			if (type(LootCount_DropCount_Character.DoneQuest[qName])=="table") then
				if (qID and qID~=true) then
					LootCount_DropCount_Character.DoneQuest[qName][qID]=true;
				end
			elseif (qID and qID~=true) then
				local num=LootCount_DropCount_Character.DoneQuest[qName];
				LootCount_DropCount_Character.DoneQuest[qName]=nil;
				LootCount_DropCount_Character.DoneQuest[qName]={};
				LootCount_DropCount_Character.DoneQuest[qName][qID]=true;
			end
		end
		LCDC_RescanQuests=LCDC_RESCANQUESTS;
		return;
	end
	if (event=="QUEST_FINISHED") then		-- Also when frame is closed (apparently)
		LCDC_RescanQuests=LCDC_RESCANQUESTS;
		return;
	end
	if (event=="UNIT_SPELLCAST_START") then

--DuckLib:Chat(" Spellcast start: "..arg2,1,0,1);

		if (arg1~="player") then return; end		-- Someone else in party
		if (arg2==LootCount_DropCount_Professions[1] or
			arg2==LootCount_DropCount_Professions[2] or
			arg2==LootCount_DropCount_Professions[3] or
			arg2==LootCount_DropCount_Professions[4]) then
			if (LootCount_DropCount_Target) then
				LootCount_DropCount_Profession=arg2;			-- Set loot-by-profession type
			end
		else
			LootCount_DropCount_Profession=nil;			-- Set normal type loot
		end
		return;
	end
	if (event=="QUEST_QUERY_COMPLETE") then
		LootCount_DropCount_DB.QuestQuery={};
		LootCount_DropCount_DB.QuestQuery=GetQuestsCompleted();
		LootCount_DropCount_GetQuestNames();
		return;
	end

	if (event=="VARIABLES_LOADED") then
		if (LootCount_DropCount_Character.ShowZoneMobs==nil) then LootCount_DropCount_Character.ShowZoneMobs=false; end
		if (LootCount_DropCount_Character.ShowZone==nil) then LootCount_DropCount_Character.ShowZone=true; end
		if (not LootCount_DropCount_DB.CHANNELSET) then
			LootCount_DropCount_DB.CHANNELSET=true;
		end
		if (LootCount_DropCount_DB.IconX and LootCount_DropCount_DB.IconY) then LCDC_Minimap_SetIconAbsolute(LootCount_DropCount_DB.IconX,LootCount_DropCount_DB.IconY);
		else
			if (not LootCount_DropCount_DB.IconPosition) then LootCount_DropCount_DB.IconPosition=180; end
			LCDC_Minimap_SetIconAngle(LootCount_DropCount_DB.IconPosition);
		end
		LCDC_MyFaction=UnitFactionGroup("player");
		LCDC_ConvertBookFormat();
		LootCount_DropCount_MakeIcons();
		LootCount_DropCount_MakeIconsBook();
		LootCount_DropCount_MakeIconsQuest();
		Astrolabe:Register_OnEdgeChanged_Callback(LootCount_DropCount_AstrolabeEdge,1);
		LootCount_DropCount_DB.RAID=nil;
		if (IsInGuild()) then LootCount_DropCount_DB.GUILD=true; else LootCount_DropCount_DB.GUILD=nil; end
		-- Merge data
		LootCount_DropCount_MergeDatabase();
		-- Clear supplied database
		LootCount_DropCount_MergeData=nil;
		LootCount_DropCount_RemoveData=nil;
		collectgarbage("collect");
		QueryQuestsCompleted();
	end
end

function LootCount_DropCount_GetQuestNames()
	if (not LootCount_DropCount_DB.QuestQuery) then return; end
	if (not LootCount_DropCount_Character.DoneQuest) then
		LootCount_DropCount_Character.DoneQuest={};
	end
	-- Remove all that is already okay
	for dqName,dqState in pairs(LootCount_DropCount_Character.DoneQuest) do
		if (type(dqState)=="table") then
			for queueNum,_ in pairs(dqState) do
				if (LootCount_DropCount_DB.QuestQuery[queueNum]) then
					LootCount_DropCount_DB.QuestQuery[queueNum]=nil;
				end
			end
		else
			if (LootCount_DropCount_DB.QuestQuery[dqState]) then
				LootCount_DropCount_DB.QuestQuery[dqState]=nil;
			end
		end
	end
	LCDC_QuestConvertItems=0;
	for _,_ in pairs(LootCount_DropCount_DB.QuestQuery) do
		LCDC_QuestConvertItems=LCDC_QuestConvertItems+1;
	end

	QuestsDone.Waiter=3;
end

function LootCount_DropCount_GetQuestName(link)
	if (not string.find(link,"quest:")) then return nil; end

	LootCount_DropCount_CF:SetOwner(WorldFrame, "ANCHOR_NONE");
	LootCount_DropCount_CF:ClearLines();
	LootCount_DropCount_CF:SetHyperlink(link);

	local text=getglobal("LootCount_DropCount_CFTextLeft1"):GetText();
	LootCount_DropCount_CF:Hide();
	return text;
end


function LCDC_GetUnitType(target)
	local uClass=UnitClassification(target);
	if (uClass=="worldboss") then uClass="World boss";
	elseif (uClass=="rareelite") then uClass="Rare elite";
	--elite
	elseif (uClass=="rare") then uClass="Rare";
	--normal
	--trivial
	else uClass=nil; end

	return uClass;
end


function LootCount_DropCount_AstrolabeEdge()
	local v,icon;
	for v,icon in pairs(LCDC_VendorIconMM) do
		if (Astrolabe:IsIconOnEdge(icon)) then
			icon:SetAlpha(.6);
		else
			icon:SetAlpha(1);
		end
	end
end


function LCDC_SaveQuest(qName,qGiver,qLevel)
	if (not qName) then return; end
	if (not qLevel) then qLevel=0; end
	if (not LCDC_LastFaction or LCDC_LastFaction=="") then LCDC_LastFaction="Neutral"; end
	if (not LootCount_DropCount_DB.Quest) then LootCount_DropCount_DB.Quest={}; end
	if (not LootCount_DropCount_DB.Quest[LCDC_LastFaction]) then LootCount_DropCount_DB.Quest[LCDC_LastFaction]={}; end
	local qTable=LootCount_DropCount_DB.Quest[LCDC_LastFaction];

	local qX,qY=LCDC_GetPLayerPosition();
	local qZone=LootCount_DropCount_GetZone();
	if (not qZone) then qZone="Unknown"; end
	if (not qGiver or qGiver=="") then qGiver="- item - ("..LootCount_DropCount_GetZone().." "..math.floor(qX)..","..math.floor(qY)..")"; end

	if (not qTable[qGiver]) then qTable[qGiver]={}; end
	qTable[qGiver].Zone=qZone;
	if (not qTable[qGiver].X or not qTable[qGiver].Y or (qTable[qGiver].X and qX~=0 and qTable[qGiver].Y and qY~=0)) then
		qTable[qGiver].X=qX;
		qTable[qGiver].Y=qY;
	end
--[[
	if (qTable[qGiver].Quests) then
		-- Convert old indexed list
		local i=1;
		while (qTable[qGiver].Quests[i] and not qTable[qGiver].Quests[qName]) do
			qTable[qGiver].Quests[qName]={ Level=0, };
			qTable[qGiver].Quests[i]=nil;
			i=i+1;
		end
	else
		qTable[qGiver].Quests={};
	end
	local newquest=nil;
	if (not qTable[qGiver].Quests[qName]) then newquest=true; end
	qTable[qGiver].Quests[qName]={ Level=qLevel, };
	if (newquest) then
		DuckLib:Chat(LCDC_C_BASIC.."New "..LCDC_LastFaction.." Quest "..LCDC_C_GREEN.."\""..qName.." ("..qLevel..")\""..LCDC_C_BASIC.." saved");
		LootCount_DropCount_MakeIconsQuest();
	end
]]
	local i=1;
	if (qTable[qGiver].Quests) then
		while (qTable[qGiver].Quests[i] and qTable[qGiver].Quests[i]~=qName) do i=i+1; end
	else
		qTable[qGiver].Quests={};
	end
	local newquest=nil;
	if (not qTable[qGiver].Quests[i]) then newquest=true; end
	qTable[qGiver].Quests[i]=qName;
	if (newquest) then
		DuckLib:Chat(LCDC_C_BASIC.."New "..LCDC_LastFaction.." Quest "..LCDC_C_GREEN.."\""..qName.."\""..LCDC_C_BASIC.." saved for "..LCDC_C_GREEN..qGiver);
--		LCDC_ScanQuests();
--		LootCount_DropCount_MakeIconsQuest();
	end

--	LootCount_DropCount_MakeIconsQuest();
--	DuckLib:Chat(LCDC_C_BASIC.."Quest "..LCDC_C_GREEN.."\""..qName.."\""..LCDC_C_BASIC.." saved");
end


function LCDC_ScanQuests()
	if (not LootCount_DropCount_Character.Quests) then LootCount_DropCount_Character.Quests={}; end
	DuckLib:ClearTable(LootCount_DropCount_Character.Quests);

	ExpandQuestHeader(0);			-- Expand all quest-headers
	local i=1;
	while (GetQuestLogTitle(i)~=nil) do
		local questTitle,level,questTag,suggestedGroup,isHeader,isCollapsed,isComplete,isDaily=GetQuestLogTitle(i);
		if (not isHeader) then
			local link=GetQuestLink(i);
			local questID,i1,i2=link:match("|Hquest:(%p?%d+):(%p?%d+)|h%[(.-)%]|h");
			LootCount_DropCount_Character.Quests[questTitle]=tonumber(questID);
		end
		i=i+1;
	end
end


function LCDC_DoesNPCHaveQuest(faction,npc,qName)
	if (not faction or not npc or not qName) then return nil; end
	if (not LootCount_DropCount_DB.Quest) then return nil; end
	if (not LootCount_DropCount_DB.Quest[faction]) then return nil; end
	if (not LootCount_DropCount_DB.Quest[faction][npc]) then return nil; end
	if (not LootCount_DropCount_DB.Quest[faction][npc].Quests) then return nil; end
	for qIndex,quest in pairs(LootCount_DropCount_DB.Quest[faction][npc].Quests) do
		if (quest==qName) then return true; end
	end
	return nil;
end


function LootCount_DropCount_MakeIconsQuest()
	if (not LCDC_Zones) then LootCount_DropCount_EnumZones(); end
	local Cont,Zone,_,_=Astrolabe:GetCurrentPlayerPosition();
	local npc,vTable;
	for npc,nTable in pairs(LCDC_QuestIconMM) do
--		nTable:Hide();
		Astrolabe:RemoveIconFromMinimap(LCDC_QuestIconMM[npc]);
	end
	if (not LootCount_DropCount_DB.Quest) then return; end
	local count=1;
	for faction,fTable in pairs(LootCount_DropCount_DB.Quest) do
		-- Check faction here

		for npc,nTable in pairs(fTable) do
			local done=nil;
			c=1;
			while(not done and LCDC_Zones[c]) do
				local z=1;
				while(not done and LCDC_Zones[c][z]) do
					if (string.find(nTable.Zone,LCDC_Zones[c][z])==1) then
						local r,g,b=0,0,0;				-- I've done it
						if (LootCount_DropCount_Character.DoneQuest) then
							for qIndex,quest in pairs(nTable.Quests) do
								if (not LootCount_DropCount_Character.DoneQuest[quest]) then
									r,g,b=0,1,0;		-- I'm not finished for some reason
									break;
								end
							end
						end
						if (LootCount_DropCount_Character.Quests) then
							for thisquest,_ in pairs(LootCount_DropCount_Character.Quests) do
								if (LCDC_DoesNPCHaveQuest(faction,npc,thisquest)) then
									r,g,b=1,1,1;		-- I have it
								end
							end
						end

						if (not LCDC_QuestIcon[npc]) then
							LCDC_QuestIcon[npc]=CreateFrame("Button","LCDC_MapQuest"..count,WorldMapDetailFrame,"LCDC_VendorFlagTemplate");
							LCDC_QuestIcon[npc].icon=LCDC_QuestIcon[npc]:CreateTexture("ARTWORK");
							LCDC_QuestIcon[npc].icon:SetTexture("Interface\\QuestFrame\\UI-Quest-BulletPoint");
							LCDC_QuestIcon[npc].icon:SetAllPoints();
						end
						LCDC_QuestIcon[npc].icon:SetVertexColor(r,g,b);
						LCDC_QuestIcon[npc].NPC={
							Name=npc,
							X=nTable.X/100,
							Y=nTable.Y/100,
							Continent=c,
							Zone=z,
							Faction=faction,
						}

						if (not LCDC_QuestIconMM[npc]) then
							LCDC_QuestIconMM[npc]=CreateFrame("Button","LCDC_MapQuestMM"..count,UIParent,"LCDC_VendorFlagTemplate");
							LCDC_QuestIconMM[npc].icon=LCDC_QuestIconMM[npc]:CreateTexture("ARTWORK");
							LCDC_QuestIconMM[npc].icon:SetTexture("Interface\\QuestFrame\\UI-Quest-BulletPoint");
							LCDC_QuestIconMM[npc].icon:SetAllPoints();
						end
						LCDC_QuestIconMM[npc].icon:SetVertexColor(r,g,b);
						LCDC_QuestIconMM[npc].NPC={
							Name=npc,
							Continent=c,
							Zone=z,
							Faction=faction,
						}

--						if (LootCount_DropCount_DB.VendorMinimap and Cont==c and Zone==z) then
						if (LootCount_DropCount_DB.QuestMinimap and Cont==c and Zone==z) then
							LootCount_DropCount_SetQuestMMPosition(npc,nTable.X,nTable.Y);
						end

						count=count+1;
						done=true;
					end
					z=z+1;
				end
				c=c+1;
			end
	--		if (not done) then
	--			DuckLib:Chat(vTable.Zone.." unmappable",0,1,1);
	--		end
		end
	end
end


function LootCount_DropCount_PlotIcons()
	if (not WorldMapDetailFrame) then return; end
	local c=GetCurrentMapContinent();
	if (c<1) then return; end		-- Cosmos or battleground or DK starting-area
	local z=GetCurrentMapZone();

	local entry,eIcon;
	for entry,eIcon in pairs(LCDC_VendorIcon) do
		if (LootCount_DropCount_DB.VendorWorldmap and eIcon.Vendor.Continent==c and eIcon.Vendor.Zone==z and (not eIcon.Vendor.Faction or eIcon.Vendor.Faction=="Neutral" or eIcon.Vendor.Faction==LCDC_MyFaction)) then
			Astrolabe:PlaceIconOnWorldMap(WorldMapDetailFrame,eIcon,c,z,eIcon.Vendor.X,eIcon.Vendor.Y);
		else
			eIcon:Hide();
		end
	end

	for entry,eIcon in pairs(LCDC_BookIcon) do
		if (LootCount_DropCount_DB.BookWorldmap and eIcon.Book.Continent==c and eIcon.Book.Zone==z) then
			Astrolabe:PlaceIconOnWorldMap(WorldMapDetailFrame,eIcon,c,z,eIcon.Book.X,eIcon.Book.Y);
		else
			eIcon:Hide();
		end
	end

	for entry,eIcon in pairs(LCDC_QuestIcon) do
--		if (LootCount_DropCount_DB.BookWorldmap and eIcon.Book.Continent==c and eIcon.Book.Zone==z) then
		if (LootCount_DropCount_DB.QuestWorldmap and eIcon.NPC.Continent==c and eIcon.NPC.Zone==z) then
			Astrolabe:PlaceIconOnWorldMap(WorldMapDetailFrame,eIcon,c,z,eIcon.NPC.X,eIcon.NPC.Y);
		else
			eIcon:Hide();
		end
	end
end


function LCDC_VendorFlag_OnEnter()
	LCDC_VendorFlag_Info=true;
	if (this.Vendor) then LootCount_DropCount_SetVendorlist(this.Vendor.Name); end
	if (this.Book) then LootCount_DropCount_SetBookTT(this.Book.Name,parent); end
	if (this.NPC) then LootCount_DropCount_SetQuestTT(this.NPC.Faction,this.NPC.Name,parent); end
end


function LootCount_DropCount_SetQuestMMPosition(npc,xPos,yPos)
	if (not LCDC_QuestIconMM[npc]) then return; end
	LCDC_QuestIconMM[npc].NPC.X=xPos/100;
	LCDC_QuestIconMM[npc].NPC.Y=yPos/100;
	Astrolabe:PlaceIconOnMinimap(LCDC_QuestIconMM[npc],LCDC_QuestIconMM[npc].NPC.Continent,LCDC_QuestIconMM[npc].NPC.Zone,LCDC_QuestIconMM[npc].NPC.X,LCDC_QuestIconMM[npc].NPC.Y);
end
function LootCount_DropCount_SetVendorMMPosition(dude,xPos,yPos)
	if (not LCDC_VendorIconMM[dude]) then return; end
	LCDC_VendorIconMM[dude].Vendor.X=xPos/100;
	LCDC_VendorIconMM[dude].Vendor.Y=yPos/100;
	Astrolabe:PlaceIconOnMinimap(LCDC_VendorIconMM[dude],LCDC_VendorIconMM[dude].Vendor.Continent,LCDC_VendorIconMM[dude].Vendor.Zone,LCDC_VendorIconMM[dude].Vendor.X,LCDC_VendorIconMM[dude].Vendor.Y);
end
function LootCount_DropCount_SetBookMMPosition(book,xPos,yPos)
	if (not LCDC_BookIconMM[book]) then return; end
	LCDC_BookIconMM[book].Book.X=xPos/100;
	LCDC_BookIconMM[book].Book.Y=yPos/100;
	Astrolabe:PlaceIconOnMinimap(LCDC_BookIconMM[book],LCDC_BookIconMM[book].Book.Continent,LCDC_BookIconMM[book].Book.Zone,LCDC_BookIconMM[book].Book.X,LCDC_BookIconMM[book].Book.Y);
end


function LootCount_DropCount_EnumZones()
	if (LCDC_Zones) then DuckLib:ClearTable(LCDC_Zones); LCDC_Zones=nil; end
	local cNames = { GetMapContinents() };
	LCDC_Zones={ };
	local c=1;
	while(cNames[c]) do
		LCDC_Zones[c]={ GetMapZones(c) } ;
		c=c+1;
	end
end


function LootCount_DropCount_ListBook(bookin)
	if (not LootCount_DropCount_DB.Book) then return; end
	local book=string.lower(bookin);
	for title,bTable in pairs(LootCount_DropCount_DB.Book) do
		if (book==string.lower(title)) then
			DuckLib:Chat(title);
			for loc,lTable in pairs(bTable) do
				DuckLib:Chat(LCDC_C_HBLUE..lTable.Zone..LCDC_C_YELLOW.." ("..string.format("%.0f,%.0f",lTable.X,lTable.Y)..")");
			end
			return;
		end
	end

	DuckLib:Chat(LCDC_C_YELLOW.."Unknown book: "..bookin.."|r");
end


function LootCount_DropCount_ListZoneBooks()
	if (not LootCount_DropCount_DB.Book) then return; end
	local found=nil;
	local here=GetRealZoneText();
	local clip=string.len(here)+3+1;
	DuckLib:Chat(here);
	for title,bTable in pairs(LootCount_DropCount_DB.Book) do
		for index,iTable in pairs(bTable) do
			if (iTable.Zone) then
				if (string.find(iTable.Zone,here)==1) then
					local subzone=string.sub(iTable.Zone,clip);
					if (not subzone) then subzone=""; end
					DuckLib:Chat(LCDC_C_HBLUE..title.." |r- "..LCDC_C_YELLOW..subzone.." ("..string.format("%.0f,%.0f",iTable.X,iTable.Y)..")");
					found=true;
				end
			end
		end
	end
	if (not found) then DuckLib:Chat(LCDC_C_YELLOW.."No known books in "..here.."|r"); end
end


function LootCount_DropCount_MakeIconsBook()
	if (not LCDC_Zones) then LootCount_DropCount_EnumZones(); end
	local Cont,Zone,_,_=Astrolabe:GetCurrentPlayerPosition();
	local book,bTable;
	for book,bTable in pairs(LCDC_BookIconMM) do
		bTable:Hide();
		Astrolabe:RemoveIconFromMinimap(LCDC_BookIconMM[book]);
	end
	if (not LootCount_DropCount_DB.Book) then return; end
	local index;
	local count=1;
	for book,vTable in pairs(LootCount_DropCount_DB.Book) do
		for index,bTable in pairs(vTable) do
			local done=nil;
			c=1;
			while(not done and LCDC_Zones[c]) do
				local z=1;
				while(not done and LCDC_Zones[c][z]) do
					if (string.find(bTable.Zone,LCDC_Zones[c][z])==1) then
						if (not LCDC_BookIcon[book..index]) then
							LCDC_BookIcon[book..index]=CreateFrame("Button","LCDC_MapBook"..count,WorldMapDetailFrame,"LCDC_VendorFlagTemplate");
							LCDC_BookIcon[book..index].icon=LCDC_BookIcon[book..index]:CreateTexture("ARTWORK");
							LCDC_BookIcon[book..index].icon:SetTexture("Interface\\Spellbook\\Spellbook-Icon");
							LCDC_BookIcon[book..index].icon:SetAllPoints();
						end
						LCDC_BookIcon[book..index].Book={
							Name=book,
							X=bTable.X/100,
							Y=bTable.Y/100,
							Continent=c,
							Zone=z,
						}

						if (not LCDC_BookIconMM[book..index]) then
							LCDC_BookIconMM[book..index]=CreateFrame("Button","LCDC_MapBookMM"..count,UIParent,"LCDC_VendorFlagTemplate");
							LCDC_BookIconMM[book..index].icon=LCDC_BookIconMM[book..index]:CreateTexture("ARTWORK");
							LCDC_BookIconMM[book..index].icon:SetTexture("Interface\\Spellbook\\Spellbook-Icon");
							LCDC_BookIconMM[book..index].icon:SetAllPoints();
						end
						LCDC_BookIconMM[book..index].Book={
							Name=book,
							Continent=c,
							Zone=z,
						}
						if (LootCount_DropCount_DB.BookMinimap and Cont==c and Zone==z) then
							LootCount_DropCount_SetBookMMPosition(book..index,bTable.X,bTable.Y);
						end

						count=count+1;
						done=true;
					end
					z=z+1;
				end
				c=c+1;
			end
--			if (not done) then
--				DuckLib:Chat(bTable.Zone.." unmappable",0,1,1);
--			end
		end
	end
end


function LootCount_DropCount_MakeIcons()
	if (not LCDC_Zones) then LootCount_DropCount_EnumZones(); end
	local Cont,Zone,_,_=Astrolabe:GetCurrentPlayerPosition();
	local vendor,vTable;
	for vendor,vTable in pairs(LCDC_VendorIconMM) do
		vTable:Hide();
		Astrolabe:RemoveIconFromMinimap(LCDC_VendorIconMM[vendor]);
	end
	if (not LootCount_DropCount_DB.Vendor) then return; end
	local count=1;
	for vendor,vTable in pairs(LootCount_DropCount_DB.Vendor) do
		local done=nil;
		c=1;
		while(not done and LCDC_Zones[c]) do
			local z=1;
			while(not done and LCDC_Zones[c][z]) do
				if (string.find(vTable.Zone,LCDC_Zones[c][z])==1) then
					if (not LCDC_VendorIcon[vendor]) then
						LCDC_VendorIcon[vendor]=CreateFrame("Button","LCDC_MapVendor"..count,WorldMapDetailFrame,"LCDC_VendorFlagTemplate");
						LCDC_VendorIcon[vendor].icon=LCDC_VendorIcon[vendor]:CreateTexture("ARTWORK");
						LCDC_VendorIcon[vendor].icon:SetTexture("Interface\\GROUPFRAME\\UI-Group-MasterLooter");
						LCDC_VendorIcon[vendor].icon:SetAllPoints();
					end
					LCDC_VendorIcon[vendor].Vendor={
						Name=vendor,
						X=vTable.X/100,
						Y=vTable.Y/100,
						Continent=c,
						Zone=z,
						Faction=vTable.Faction,
					}

					if (not LCDC_VendorIconMM[vendor]) then
						LCDC_VendorIconMM[vendor]=CreateFrame("Button","LCDC_MapVendorMM"..count,UIParent,"LCDC_VendorFlagTemplate");
						LCDC_VendorIconMM[vendor].icon=LCDC_VendorIconMM[vendor]:CreateTexture("ARTWORK");
						LCDC_VendorIconMM[vendor].icon:SetTexture("Interface\\GROUPFRAME\\UI-Group-MasterLooter");
						LCDC_VendorIconMM[vendor].icon:SetAllPoints();
					end
					LCDC_VendorIconMM[vendor].Vendor={
						Name=vendor,
						Continent=c,
						Zone=z,
						Faction=vTable.Faction,
					}
					if (LootCount_DropCount_DB.VendorMinimap and Cont==c and Zone==z) then
						LootCount_DropCount_SetVendorMMPosition(vendor,vTable.X,vTable.Y)
					end

					count=count+1;
					done=true;
				end
				z=z+1;
			end
			c=c+1;
		end
--		if (not done) then
--			DuckLib:Chat(vTable.Zone.." unmappable",0,1,1);
--		end
	end
end


function LootCount_DropCount_FindByName(name)
	if (not LootCount_DropCount_DB.Vendor) then return nil; end
	name=string.lower(name);

	for vendor,vTable in pairs(LootCount_DropCount_DB.Vendor) do
		if (vTable.Items) then
			for item,iTable in pairs(vTable.Items) do
				if (iTable.Name and name==string.lower(iTable.Name)) then
					return item;
				end
			end
		end
	end
	return nil;
end


function LootCount_DropCount_ConvertDB_MK2()
	if (LootCount_DropCount_DB.Converted==2) then return; end
	for id,idValue in pairs(LootCount_DropCount_SkinningList) do
		local dbItem="item:"..id..":0:0:0:0:0:0";
		if (LootCount_DropCount_DB.Item[dbItem] and LootCount_DropCount_DB.Item[dbItem].Name) then
			LootCount_DropCount_DB.Item[dbItem].Skinning=DuckLib:CopyTable(LootCount_DropCount_DB.Item[dbItem].Name);
			LootCount_DropCount_DB.Item[dbItem].Name=nil;
		end
	end
	LootCount_DropCount_DB.Converted=2;
end


function LootCount_DropCount_ConvertDB()
	if (not LootCount_DropCount_DB.Item) then LootCount_DropCount_DB.Item={}; end
	if (not LootCount_DropCount_DB.Vendor) then LootCount_DropCount_DB.Vendor={}; end
	if ((#LootCount_DropCount_DB.Item)==0 and (#LootCount_DropCount_DB.Vendor)==0) then
		LootCount_DropCount_DB.Converted=true; LCDC_MaintainerSPV=SP_VendorLocations; LCDC_MaintainerSPI=SP_VendorInfo;
	end
	if (LootCount_DropCount_DB.Converted) then
		if (LootCount_DropCount_DB.Converted==true) then LootCount_DropCount_ConvertDB_MK2(); end
		return;
	end
	if (LCDC_CachedConvertItems==0) then
		StaticPopupDialogs["LCDC_D_NOTIFICATION"].text=LCDC_C_BASIC.."DropCount\n\n"..LCDC_C_YELLOW;
		StaticPopupDialogs["LCDC_D_NOTIFICATION"].text=StaticPopupDialogs["LCDC_D_NOTIFICATION"].text..[[
This DropCount version requires the database to be converted to the new format.]];
		StaticPopupDialogs["LCDC_D_NOTIFICATION"].text=StaticPopupDialogs["LCDC_D_NOTIFICATION"].text..LCDC_C_GREEN;
		StaticPopupDialogs["LCDC_D_NOTIFICATION"].text=StaticPopupDialogs["LCDC_D_NOTIFICATION"].text.."\nYou can play as normal while this is done.\n"..LCDC_C_YELLOW;
		StaticPopupDialogs["LCDC_D_NOTIFICATION"].text=StaticPopupDialogs["LCDC_D_NOTIFICATION"].text..[[
The conversion operation may take a long time.
If you log out during the process, it will continue where it left off when you log back in.]];
		StaticPopup_Show("LCDC_D_NOTIFICATION");
		DuckLib:Chat("Conversion in progress...",1,0,0);
	end

	local item,iTable;
	LCDC_CachedConvertItems=0;
	if (LootCount_DropCount_DB.Item) then
		for item,iTable in pairs(LootCount_DropCount_DB.Item) do
			if (not iTable.Item) then					-- Conversion needed
				local itemName,itemLink=GetItemInfo(item);
				if (not itemName) then
					LootCount_DropCount_StartCache(item);
					LCDC_CachedConvertItems=LCDC_CachedConvertItems+1;
				else
					iTable.Item=itemName;
				end
			end
		end
	end

	if (LootCount_DropCount_DB.Vendor) then
		local vendor,vTable;
		for vendor,vTable in pairs(LootCount_DropCount_DB.Vendor) do
			for item,iTable in pairs(LootCount_DropCount_DB.Vendor[vendor].Items) do
				if (type(iTable)=="number") then		-- Conversion needed
					local itemName,itemLink=GetItemInfo(item);
					if (not itemName) then
						LootCount_DropCount_StartCache(item);
						LCDC_CachedConvertItems=LCDC_CachedConvertItems+1;
					else
						local count=LootCount_DropCount_DB.Vendor[vendor].Items[item];
						LootCount_DropCount_DB.Vendor[vendor].Items[item]={};
						LootCount_DropCount_DB.Vendor[vendor].Items[item].Name=itemName;
						LootCount_DropCount_DB.Vendor[vendor].Items[item].Count=count;
					end
				end
			end
		end
	end

	if (LCDC_CachedConvertItems==0) then
		LootCount_DropCount_DB.Converted=true;
		DuckLib:Chat("DropCount database conversion done.",0,1,0);
		collectgarbage("collect");
	else
		DuckLib:Chat(LCDC_CachedConvertItems.." items to read from server for converting the database.",1,0,0);
	end
end


function LootCount_DropCount_ItemVendors()
	if (not LootCount_DropCount_QueryItem) then return; end

	local itemName,itemLink=GetItemInfo(LootCount_DropCount_QueryItem);
	if (itemLink) then
		local itemID=DuckLib:GetID(itemLink);
		local list=LootCount_DropCount_TooltipVendor(itemID,true);

		-- Type list
		local line=1;
		while(list[line]) do
			if (line==1) then DuckLib:Chat(itemLink); end
			DuckLib:Chat(list[line].Ltext.." "..list[line].Rtext);
			line=line+1;
		end

		if (line==1) then
			DuckLib:Chat(LCDC_C_YELLOW.."No known vendors for "..itemName.."|r");
		end
	end
end

-- MerchantRepairAllButton
function LootCount_DropCount_ReadMerchant(dude)
	local rebuildIcons=nil;
	local numItems=GetMerchantNumItems();
	if (not numItems or numItems<1) then return true; end
	if (not LootCount_DropCount_DB.Vendor) then LootCount_DropCount_DB.Vendor={}; end

	if (not LootCount_DropCount_DB.Vendor[dude]) then
		LootCount_DropCount_DB.Vendor[dude]={ Items={} };
		rebuildIcons=true;
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_GREEN.."New vendor added to database|r");
	end

	local posX,posY=LCDC_GetPLayerPosition();
	LootCount_DropCount_DB.Vendor[dude].Repair=_G.MerchantRepairAllButton:IsVisible();
	LootCount_DropCount_DB.Vendor[dude].Zone=LootCount_DropCount_GetZone();
	LootCount_DropCount_DB.Vendor[dude].X=posX;
	LootCount_DropCount_DB.Vendor[dude].Y=posY;
	LootCount_DropCount_DB.Vendor[dude].Faction=LCDC_LastFaction;
	LootCount_DropCount_SetVendorMMPosition(dude,posX,posY);

	-- Remove all permanent items
	for item,avail in pairs(LootCount_DropCount_DB.Vendor[dude].Items) do
		if (not avail or not avail.Count or avail.Count==LCDC_PERMANENTITEM) then
			LootCount_DropCount_DB.Vendor[dude].Items[item]=nil;
		end
	end

	-- Add all items
	local ReadOk=true;
	local index=1;
	while (index<=numItems) do
		local link=GetMerchantItemLink(index);
		link=DuckLib:GetID(link);
		if (link) then
			local itemName,itemLink=GetItemInfo(link);
			local _,_,_,_,count=GetMerchantItemInfo(index);			-- count==-1 unlimited
			LootCount_DropCount_DB.Vendor[dude].Items[link]={};		-- Create
			LootCount_DropCount_DB.Vendor[dude].Items[link].Name=itemName;
			LootCount_DropCount_DB.Vendor[dude].Items[link].Count=count;
		else
			ReadOk=nil;
		end
		index=index+1;
	end
	if (not ReadOk) then
		if (not LootCount_DropCount_VendorProblem) then
			DuckLib:Chat("Unchached item(s) at this vendor. Look through the vendor-pages to load missing items from the server.",1);
			LootCount_DropCount_VendorProblem=true;
		end
	else
		if (LootCount_DropCount_VendorProblem) then
			DuckLib:Chat("Vendor saved",0,1,0);
			rebuildIcons=true;
		end
		if (rebuildIcons) then LootCount_DropCount_MakeIcons(); end
	end
	return ReadOk;
end


function LootCount_DropCount_TransmitData(GUID,mob,item,count)
	local channel;
	if (LootCount_DropCount_DB.GUILD) then channel="GUILD";
	elseif (LootCount_DropCount_DB.RAID) then channel="RAID";
	else return; end

	local text;
	if (item and count) then
		text=LCDC_MOBLOOT..LCDC_SEPARATOR..GUID..LCDC_SEPARATOR..mob..LCDC_SEPARATOR..item..LCDC_SEPARATOR..count;
	else
		text=LCDC_MOBKILL..LCDC_SEPARATOR..GUID..LCDC_SEPARATOR..mob;
	end

	SendAddonMessage(LCDC_PREFIX,text,channel);
end


-- LCDC_MOBKILL - GUID - MOBNAME - ZONE
-- LCDC_MOBLOOT - GUID - MOBNAME - ITEM - COUNT
function LootCount_DropCount_ParseMessage(text,sender)
	local header,guid,mob,item,count=strsplit(LCDC_SEPARATOR,text);
	if (header==LCDC_MOBKILL) then
		LootCount_DropCount_AddKill(nil,guid,mob,nil,nil,true,item);
		if (LootCount_DropCount_Debug) then
			DuckLib:Chat(sender.." kill: \'"..mob.."\'",0,1,0);
		end
	elseif (header==LCDC_MOBLOOT) then
		if (not count or not item) then return; end
		if (LootCount_DropCount_LootIsSkinning(item)) then
			if (not LootCount_DropCount_AlreadyReceivedSkin(guid)) then
				LootCount_DropCount_SkinTargetKill=mob;
			end
		end
		LootCount_DropCount_AddLoot(guid,mob,item,count,true);
		local itemname=GetItemInfo(item);
		if (not itemname) then _,itemname=DuckLib:GetID(item); end
		if (LootCount_DropCount_Debug) then
			DuckLib:Chat(sender.." drop: \'"..mob.."\' -> \'"..itemname.."\'x"..count,0,1,0);
		end
	end
end


function LootCount_DropCount_AlreadyReceivedSkin(GUID)
	if (LootCount_DropCount_ReceivedSkins.A~=GUID and
		LootCount_DropCount_ReceivedSkins.B~=GUID and
		LootCount_DropCount_ReceivedSkins.C~=GUID) then
		LootCount_DropCount_ReceivedSkins.C=LootCount_DropCount_ReceivedSkins.B;
		LootCount_DropCount_ReceivedSkins.B=LootCount_DropCount_ReceivedSkins.A;
		LootCount_DropCount_ReceivedSkins.A=GUID;
		return nil;
	end
	return true;
end


function LootCount_DropCount_GetZone()
	local here=GetRealZoneText();			-- Set zone for last kill
	if (not here or here=="") then return nil; end
	local ss=GetSubZoneText();										-- Set subzone/area for last kill
	if (ss=="") then ss=nil; end
	if (ss) then here=here.." - "..ss; end
	return here;
end


function LCDC_GetPLayerPosition()
	local posX,posY=GetPlayerMapPosition("player");
	posX=(floor(posX*100000))/1000;
	posY=(floor(posY*100000))/1000;
	if (posX==0 or posY==0) then
		DuckLib:Chat("Invalid player position",1);
	end
	return posX,posY;
end


function LootCount_DropCount_AddKill(oma,GUID,mob,reservedvariable,noadd,notransmit,otherzone)
	-- Check if already counted
	local i=LootCount_DropCount_QueueSize;
	while (i>0) do
		if (LootCount_DropCount_TimedQueue[i].GUID==GUID) then return; end
		i=i-1;
	end

	local now=time();
	if (not LootCount_DropCount_DB.Count[mob]) then LootCount_DropCount_DB.Count[mob]={ Kill=0 }; end
	if (not otherzone) then otherzone=LootCount_DropCount_GetZone(); end
	LootCount_DropCount_DB.Count[mob].Zone=otherzone;		-- Set zone for last kill
	if (not noadd) then
		LootCount_DropCount_DB.Count[mob].Kill=LootCount_DropCount_DB.Count[mob].Kill+1;
		if (not notransmit) then LootCount_DropCount_TransmitData(GUID,mob); end
	else
		return;
	end

	LootCount_DropCount_QueueSize=LootCount_DropCount_QueueSize+1;
	LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize]={};
	i=LootCount_DropCount_QueueSize;
	while (i>1) do
		LootCount_DropCount_TimedQueue[i].Mob=LootCount_DropCount_TimedQueue[i-1].Mob;
		LootCount_DropCount_TimedQueue[i].GUID=LootCount_DropCount_TimedQueue[i-1].GUID;
		LootCount_DropCount_TimedQueue[i].Oma=LootCount_DropCount_TimedQueue[i-1].Oma;
		LootCount_DropCount_TimedQueue[i].Time=LootCount_DropCount_TimedQueue[i-1].Time;
		i=i-1;
	end
	LootCount_DropCount_TimedQueue[1].Mob=mob;
	LootCount_DropCount_TimedQueue[1].GUID=GUID;
	LootCount_DropCount_TimedQueue[1].Oma=oma;
	LootCount_DropCount_TimedQueue[1].Time=now;
end


function LootCount_DropCount_LootIsSkinning(item)
	if (type(item)=="string") then _,item=DuckLib:GetID(item); end
	return LootCount_DropCount_SkinningList[item];
end


function LootCount_DropCount_AddLocalLoot(GUID,mob,item,count,notransmit)
	if (not LootCount_DropCount_Character.Skinning) then
		LootCount_DropCount_Character.Skinning=LootCount_DropCount_LootIsSkinning(item);			-- Update the skinning-flag
	end
	LootCount_DropCount_AddLoot(GUID,mob,item,count,notransmit,LootCount_DropCount_Character.Skinning);
end


function LootCount_DropCount_AddLootMob(GUID,mob,item)
	local nameTable;
	if (LootCount_DropCount_Profession) then
		if (not LootCount_DropCount_DB.Item[item].Skinning) then LootCount_DropCount_DB.Item[item].Skinning={}; end
		nameTable=LootCount_DropCount_DB.Item[item].Skinning;
	else
		if (not LootCount_DropCount_DB.Item[item].Name) then LootCount_DropCount_DB.Item[item].Name={}; end
		nameTable=LootCount_DropCount_DB.Item[item].Name;
	end
	-- New stuff, so make database ready
	if (not nameTable[mob]) then		-- Mob not in drop-list
		nameTable[mob]=0;				-- Not looted, but make entry (will be added later)
	end
	if (not LootCount_DropCount_DB.Count[mob]) then					-- Mob not in kill-list
		LootCount_DropCount_AddKill(nil,GUID,mob,nil,true,true,"");	-- Add it with zero kills
	end
end


-- "isskinner" is 'true' if the loot originator is a skinner, and 'nil' if not.
-- The origin of the data can be both communication and own loot.
--
-- As of now, this function does not use this information
function LootCount_DropCount_AddLoot(GUID,mob,item,count,notransmit,isskinner)
	if (not notransmit) then LootCount_DropCount_TransmitData(GUID,mob,item,count); end
	local now=time();
	if (not LootCount_DropCount_DB.Item[item]) then LootCount_DropCount_DB.Item[item]={ }; end
	local itemName,itemLink=GetItemInfo(item);
	LootCount_DropCount_DB.Item[item].Item=itemName;
	LootCount_DropCount_DB.Item[item].Time=now;				-- Last point in time for loot of this item
	LootCount_DropCount_AddLootMob(GUID,mob,item);			-- Make register
	local skinning=nil;
	local nameTable;
	if (LootCount_DropCount_Profession) then
		nameTable=LootCount_DropCount_DB.Item[item].Skinning;
		skinning=true;
	else
		nameTable=LootCount_DropCount_DB.Item[item].Name;
	end

	nameTable[mob]=nameTable[mob]+count;

	if (skinning) then		-- Skinner-loot, so add it as a skinning-kill
		if (LootCount_DropCount_SkinTargetKill and LootCount_DropCount_SkinTargetKill==mob) then
			if (not LootCount_DropCount_DB.Count[mob].Skinning) then LootCount_DropCount_DB.Count[mob].Skinning=0; end
			LootCount_DropCount_DB.Count[mob].Skinning=LootCount_DropCount_DB.Count[mob].Skinning+1;
			LootCount_DropCount_SkinTargetKill=nil;					-- Added, so next loot on this target is more than one items from same skinning
		end
	end
	if (LootCount_DropCount_Registered) then LootCountAPI.Force(LOOTCOUNT_DROPCOUNT); end
end


function LootCount_DropCount_GetRatio(item,mob)
	if (LCT_QUEST) then
		local _,_,_,_,_,itemtype=GetItemInfo(item);
		local _,itemID=DuckLib:GetID(item);
		if (itemtype and itemtype==LCT_QUEST and not LootCount_DropCount_NoQuest[itemID]) then return LCR_QUEST,LCR_QUEST; end
	end

	local nosafe=nil;
	local nKills,nRatio=0,0;
	local sKills,sRatio=0,0;
	if (not LootCount_DropCount_DB.Item[item]) then return 0,0,true; end
	if (not LootCount_DropCount_DB.Item[item].Name and
		not LootCount_DropCount_DB.Item[item].Skinning) then return 0,0,true; end
	if (LootCount_DropCount_DB.Item[item].Name and not LootCount_DropCount_DB.Item[item].Name[mob]) then
		nRatio=0;
	end
	if (LootCount_DropCount_DB.Item[item].Skinning and not LootCount_DropCount_DB.Item[item].Skinning[mob]) then
		sRatio=0;
	end
	if (not LootCount_DropCount_DB.Count[mob]) then return 0,0,true; end

	if (LootCount_DropCount_DB.Item[item].Name) then
		nKills=LootCount_DropCount_DB.Count[mob].Kill;
		if (not nKills or nKills<1) then nRatio=0;
		else
			if (LootCount_DropCount_DB.Item[item].Name[mob]) then
				nRatio=LootCount_DropCount_DB.Item[item].Name[mob]/nKills;
				if (LootCount_DropCount_DB.Item[item].Name[mob]<2) then unsafe=true; end
			else
				nRatio=0;
			end
		end
	end
	if (LootCount_DropCount_DB.Item[item].Skinning) then
		sKills=LootCount_DropCount_DB.Count[mob].Skinning;
		if (not sKills or sKills<1) then sRatio=0;
		else
			if (LootCount_DropCount_DB.Item[item].Skinning[mob]) then
				sRatio=LootCount_DropCount_DB.Item[item].Skinning[mob]/sKills;
				if (LootCount_DropCount_DB.Item[item].Skinning[mob]<2) then unsafe=true; end
			else
				sRatio=0;
			end
		end
	end

	return nRatio,sRatio,nosafe;
end


function LootCount_DropCount_UpdateButton(button)
	if (not button) then return; end			-- End of iteration

	if (not button.User or not button.User.Texture) then return; end
	local texture=getglobal(button:GetName().."IconTexture");
	texture:SetTexture(button.User.Texture);			-- Set texture from item
	if (not LootCount_DropCount_DB.Item[button.User.itemID]) then return; end			-- Nothing assigned yet
	if (not LootCount_DropCount_DB.Item[button.User.itemID].Name and
		not LootCount_DropCount_DB.Item[button.User.itemID].Skinning) then return; end	-- No known droppers

	local inqueue={};
	local i=1;
	while (i<=LootCount_DropCount_QueueSize and LootCount_DropCount_TimedQueue[i]) do
		if (LootCount_DropCount_TimedQueue[i].Oma) then
			if (not inqueue[LootCount_DropCount_TimedQueue[i].Mob]) then
				local drop,sD=LootCount_DropCount_GetRatio(button.User.itemID,LootCount_DropCount_TimedQueue[i].Mob);
				if (not drop or drop==0) then drop=sD; end
				inqueue[LootCount_DropCount_TimedQueue[i].Mob]={ Count=1, Ratio=drop };
			else
				inqueue[LootCount_DropCount_TimedQueue[i].Mob].Count=inqueue[LootCount_DropCount_TimedQueue[i].Mob].Count+1;
			end
		end
		i=i+1;
	end

	local ratio=0;
	local count=0;
	for mob,mTable in pairs(inqueue) do
		count=count+mTable.Count;
		ratio=ratio+(mTable.Count*mTable.Ratio);
	end
	if (count<1) then return; end
	ratio=ratio/count;

	local goalvalue=nil;
	if (button.goal and button.goal>0) then
		local amount=LootCount_GetItemCount(button.User.itemID);
		if (amount>=button.goal) then goalvalue="OK";
		else
			goalvalue=button.goal-amount;
			if (ratio>0) then goalvalue=math.ceil(goalvalue/ratio);
			else goalvalue=""; end
		end
	end

	if (LootCount_DropCount_Registered) then LootCountAPI.SetData(LOOTCOUNT_DROPCOUNT,button,LootCount_DropCount_FormatPC(ratio),goalvalue); end
end


function LootCount_DropCount_FormatPC(ratio,addition)
	if (ratio<0) then return "Quest"; end
	if (not addition) then addition=""; end
	local text;
	local pc=ratio*100;
	if (pc>=100) then text=string.format("%.0f",pc);
	elseif (pc>=10) then text=string.format("%.0f",pc);
	elseif (pc>=1) then text=string.format("%.1f",pc);
	elseif (pc==0) then return "0"..addition;
	else text=string.format("%.02f",pc);
	end
	return text..addition;
end


function LootCount_DropCount_MakeRegister(button,itemID,clearit)
	if (not button.User) then button.User = { };
	elseif (clearit) then DuckLib:ClearTable(button.User);
	end
	if (not button.User.itemID or itemID) then
		if (not itemID) then return nil; end
		button.User.itemID=itemID;
	end
	return true;
end


function LootCount_DropCount_DropItem(button,itemID)
	LootCount_DropCount_MakeRegister(button,itemID,true);						-- Start counting from now
	_,_,_,_,_,_,_,_,_,button.User.Texture=GetItemInfo(itemID);					-- Set custom texture
end


function LootCount_DropCount_GetTargetType()
	local targettype="playertarget";
	local mobname=UnitName(targettype);
	if (not mobname) then
		mobname=UnitName("focus"); if (not mobname) then return nil; end
		targettype="focus";
	end
	return targettype;
end


function LootCount_DropCount_IconClicked(button,LR,count)
	if (count~=1) then return; end
	if (LR=="RightButton") then
		LootCount_DropCount_ToggleMenu(button);
	end
end


function LootCount_DropCount_GetColour(ratio)
	if (ratio>1) then ratio=1; elseif (ratio<0) then ratio=0; end
	ratio=string.format("|cFF%02X%02X%02X",128+(ratio*127),128+(ratio*127),128+(ratio*127));		-- AARRGGBB
	return ratio;
end


function LootCount_DropCount_TooltipVendor(button,getlist)
	if (not LootCount_DropCount_DB.Converted) then
		DuckLib:Chat(LCDC_C_RED.."The DropCount database is currently being converted to the new format. Your data will be available when this is done.|r");
		return;
	end

	if (type(button)=="string") then
		button={
			FreeFloat=true;
			User={
				itemID=button;
			};
		};
	elseif (not button.User or not button.User.itemID) then
		return;
	end
	if (not LootCount_DropCount_DB.Vendor) then LootCount_DropCount_DB.Vendor={}; end
	local itemname,_,rarity=GetItemInfo(button.User.itemID);
	if (not itemname or not rarity) then LootCount_DropCount_StartCache(button.User.itemID) return; end
	local _,_,_,colour=GetItemQualityColor(rarity);
	if (not getlist) then
		if (button.FreeFloat) then GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
		else GameTooltip:SetOwner(button,"ANCHOR_RIGHT"); end
		GameTooltip:SetText(colour.."["..itemname.."]|r");
	end
	local currentzone=GetRealZoneText();
	if (not currentzone) then
		currentzone="";
	elseif (LootCount_DropCount_Character.ShowZoneMobs) then
		if (not getlist) then
			GameTooltip:AddDoubleLine("Showing vendors from "..currentzone.." only","",0,1,1,0,1,1);
		end
	end

	local ThisZone=GetRealZoneText();
	local list={};
	local line=1;
	local droplist=0;
	for vendor,vTable in pairs(LootCount_DropCount_DB.Vendor) do
		if (vTable.Items and vTable.Items[button.User.itemID]) then
			list[line]={};
			local zone=LCDC_C_LBLUE;
			if (vTable.Zone and string.find(vTable.Zone,ThisZone)==1) then zone=LCDC_C_HBLUE; end
			zone=zone..vTable.Zone.." - "..floor(vTable.X)..","..floor(vTable.Y).."|r";
			list[line].Ltext=zone.." : "..LCDC_C_YELLOW..vendor.."|r ";
			list[line].Rtext="";
			if (type(vTable.Items[button.User.itemID])=="table") then
				if (vTable.Items[button.User.itemID].Count>=0) then list[line].Rtext=LCDC_C_RED.."*|r";
				elseif (vTable.Items[button.User.itemID].Count==LCDC_UNKNOWNCOUNT) then list[line].Rtext=LCDC_C_YELLOW.."*|r";
				end
				if (LootCount_DropCount_Character.ShowZoneMobs and vTable.Zone and currentzone and string.find(vTable.Zone,currentzone)~=1) then
					list[line]=nil;
					line=line-1;
				end
			end
			line=line+1;
		end
	end

	list=LootCount_DropCount_SortNames(list);
	if (getlist) then return DuckLib:CopyTable(list); end
	if (line==1) then
		GameTooltip:AddLine("No known vendors",1,1,1);
	else
		-- Type list
		line=1;
		while(list[line]) do
			GameTooltip:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
			line=line+1;
		end
	end

	GameTooltip:Show();
end


function LootCount_DropCount_Tooltip(button)
	if (type(button)=="string") then
		button={
			FreeFloat=true;
			User={
				itemID=button;
			};
		};
	elseif (not button.User or not button.User.itemID) then
		GameTooltip:SetOwner(button,"ANCHOR_RIGHT");
		GameTooltip:SetText("Drop an item here");
		GameTooltip:Show();
		return;
	end
	if (not LootCount_DropCount_DB.Item[button.User.itemID]) then return; end
	local itemname,_,rarity=GetItemInfo(button.User.itemID);
	if (not itemname or not rarity) then LootCount_DropCount_StartCache(button.User.itemID) return; end
	local _,_,_,colour=GetItemQualityColor(rarity);
	if (button.FreeFloat) then GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
	else GameTooltip:SetOwner(button,"ANCHOR_RIGHT"); end
	GameTooltip:SetText(colour.."["..itemname.."]|r:");
	local currentzone=GetRealZoneText();
	if (not currentzone) then
		currentzone="";
	elseif (LootCount_DropCount_Character.ShowZoneMobs) then
		GameTooltip:AddDoubleLine("Showing mobs from "..currentzone.." only","",0,1,1,0,1,1);
	end
	local skinningdrop=LootCount_DropCount_DB.Item[button.User.itemID].Skinning;
	local normaldrop=LootCount_DropCount_DB.Item[button.User.itemID].Name;
	if (skinningdrop) then
		if (normaldrop) then GameTooltip:AddDoubleLine("Loot and |cFFFF00FFprofession","",1,1,1,1,1,1);
		else GameTooltip:AddDoubleLine("Profession","",1,0,1,1,0,1); end
	end

	-- Init raw list
	local list={};
	local line=1;
	local droplist=0;

	-- Do normal loot
	if (normaldrop) then
		for mob,drops in pairs(LootCount_DropCount_DB.Item[button.User.itemID].Name) do
			local pretext="";
			local i=1;
			local Show=nil;
			while(LootCount_DropCount_TimedQueue[i]) do
				if (LootCount_DropCount_TimedQueue[i].Mob==mob and LootCount_DropCount_TimedQueue[i].Oma) then
					pretext="-> ";
					Show=true;							-- Show this one no matter what
					droplist=droplist+1;				-- Count a hi-pri entry
					i=LootCount_DropCount_QueueSize;	-- Break
				end
				i=i+1;
			end

			local low,high=64,255;
			list[line]={};

			if (LootCount_DropCount_DB.Count[mob].Kill) then
				list[line].Count=LootCount_DropCount_DB.Count[mob].Kill;
			else
				list[line].Count=0;
			end
			local saturation=((high-low)/(LCDC_KILLS_SAFE-LCDC_KILLS_HIDE))*list[line].Count;
			if (saturation<0) then saturation=0;
			elseif (saturation>(high-low)) then saturation=(high-low); end
			local colour=string.format("|cFF%02X%02X%02X",high-saturation,low+saturation,0);			-- AARRGGBB
			list[line].ratio=LootCount_DropCount_GetRatio(button.User.itemID,mob);	-- Normal

			local zone="";
			if (LootCount_DropCount_Character.ShowZone and LootCount_DropCount_DB.Count[mob].Zone) then
				zone=" |cFF0060FF("..LootCount_DropCount_DB.Count[mob].Zone..")";
			end

			list[line].Ltext=colour..pretext..mob..zone.."|r: ";
			list[line].Show=Show;
			if (LootCount_DropCount_Character.ShowZoneMobs and LootCount_DropCount_DB.Count[mob].Zone and currentzone and string.find(LootCount_DropCount_DB.Count[mob].Zone,currentzone)~=1) then
				list[line]=nil;
				line=line-1;
			end
			line=line+1;
		end
	end

	-- Do profession-loot
	if (skinningdrop) then
		for mob,drops in pairs(LootCount_DropCount_DB.Item[button.User.itemID].Skinning) do
			local pretext="";
			local i=1;
			local Show=nil;
			while(LootCount_DropCount_TimedQueue[i]) do
				if (LootCount_DropCount_TimedQueue[i].Mob==mob and LootCount_DropCount_TimedQueue[i].Oma) then
					pretext="-> ";
					Show=true;							-- Show this one no matter what
					droplist=droplist+1;				-- Count a hi-pri entry
					i=LootCount_DropCount_QueueSize;	-- Break
				end
				i=i+1;
			end

			local low,high=64,255;
			list[line]={};

			if (LootCount_DropCount_DB.Count[mob].Skinning) then
				list[line].Count=LootCount_DropCount_DB.Count[mob].Skinning;
			else
				list[line].Count=0;
			end
			local saturation=((high-low)/(LCDC_KILLS_SAFE-LCDC_KILLS_HIDE))*list[line].Count;
			if (saturation<0) then saturation=0;
			elseif (saturation>(high-low)) then saturation=(high-low); end
			local colour=string.format("|cFF%02X%02X%02X",high-saturation,low+saturation,0);			-- AARRGGBB
			_,list[line].ratio=LootCount_DropCount_GetRatio(button.User.itemID,mob); -- Skinning

			local zone="";
			if (LootCount_DropCount_Character.ShowZone and LootCount_DropCount_DB.Count[mob].Zone) then
				zone=" |cFF0060FF("..LootCount_DropCount_DB.Count[mob].Zone..")";
			end

			list[line].Ltext=colour..pretext..mob..zone.."|r";
			if (normaldrop) then list[line].Ltext="|cFFFF00FF*|r "..list[line].Ltext; end
			list[line].Show=Show;
			if (LootCount_DropCount_Character.ShowZoneMobs and LootCount_DropCount_DB.Count[mob].Zone and currentzone and string.find(LootCount_DropCount_DB.Count[mob].Zone,currentzone)~=1) then
				list[line]=nil;
				line=line-1;
			end
			line=line+1;
		end
	end
	list=LootCount_DropCount_SortRatio(list);

	-- Type list
	local count=0;
	local supressed=0;
	local goal=LCDC_LISTLENGTH-droplist;						-- Subtract hi-pri entry
	line=1;
	while(list[line]) do
		if ((count<goal and list[line].Count>=LCDC_KILLS_HIDE) or list[line].Show) then
			GameTooltip:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
			if (not list[line].Show) then count=count+1; end	-- Count all normal entries
		else
			supressed=supressed+1;
		end
		line=line+1;
	end
	if (supressed>0) then
		GameTooltip:AddDoubleLine(supressed.." more entries","",1,.5,1,1,1,1);
	end

	GameTooltip:Show();
end


function LootCount_DropCount_SwapListTables(thelist,A,B)
	local storage=DuckLib:CopyTable(thelist[A]);
	thelist[A]=DuckLib:CopyTable(thelist[B]);
	thelist[B]=DuckLib:CopyTable(storage);
end


function LootCount_DropCount_SortRatio(list)
	local line=1;
	while(list[line+1]) do
		if (list[line].ratio<list[line+1].ratio) then
			LootCount_DropCount_SwapListTables(list,line,line+1);
			if (line>1) then line=line-2; end
		end
		line=line+1;
	end
	local line=1;
	while(list[line]) do
		local addin=LootCount_DropCount_FormatPC(list[line].ratio,"%|r");
		if (list[line].NoSafe) then addin="<"..addin; end
		list[line].Rtext=LootCount_DropCount_GetColour(list[line].ratio)..addin;
		line=line+1;
	end
	return list;
end


function LootCount_DropCount_SortNames(list)
	local line=1;
	while(list[line+1]) do
		if (list[line].Ltext>list[line+1].Ltext) then
			LootCount_DropCount_SwapListTables(list,line,line+1);
			if (line>1) then line=line-2; end
		end
		line=line+1;
	end
	return list;
end


function LootCount_DropCount_ScanQuests()
	ExpandQuestHeader(0);			-- Expand all quest-headers
	local lastHeader=nil;
	local i=1;
	local count=0;
	while (GetQuestLogTitle(i)~=nil) do
		local questTitle,level,questTag,suggestedGroup,isHeader,isCollapsed,isComplete,isDaily=GetQuestLogTitle(i);
		if (isHeader) then
			lastHeader=questTitle;
		else
			count=count+1;
			SelectQuestLogEntry(i);
			if (not LootCount_Quest.List[questTitle]) then LootCount_Quest.List[questTitle]={}; end
			LootCount_Quest.List[questTitle]["<SECTION>"]=lastHeader;
			local goal=GetNumQuestLeaderBoards();
			if (goal and goal>0) then
				local peek=1;
				local itemIndex=1;
				while(peek<=goal) do
					local desc,oType,done=GetQuestLogLeaderBoard(peek);
					if (oType=="item") then
						local _,_,itemName,numItems,numNeeded=string.find(desc, "(.*):%s*([%d]+)%s*/%s*([%d]+)");
--						LootCount_Quest.List[questTitle][itemIndex]={ Name=itemName, Count=tonumber(numNeeded), };
						if (itemName) then
							LootCount_Quest.List[questTitle][itemName]=tonumber(numNeeded);
						end
						itemIndex=itemIndex+1;
					end
					peek=peek+1;
				end
			end
		end
		i=i+1;
	end

	-- Make the distilled item-list
	LootCount_Quest.ItemList=nil;
	LootCount_Quest.ItemList={};
	for quest,qTable in pairs(LootCount_Quest.List) do
		for item,iEntry in pairs(qTable) do
			if (item~="<SECTION>") then
				LootCount_Quest.ItemList[item]=true;					-- Can display if needed
			end
		end
	end

	-- For debug purposes
--	LootCount_Chat("Found "..count.." quests");
--	LootCount_Populated.Quest=LootCount_CopyTable(LootCount_Quest.List);
end


function LootCount_DropCount_GetQuestStatus(checkquest)
	local White="|cFFFFFFFF";
	local Yellow="|cFFFFFF00";
	local Red="|cFFFF0000";
	local Green="|cFF00FF00";
	local bBlue="|cFFA0A0FF";		-- Bright blue
	local Dark="|cFF808080";

	if (not checkquest) then return LCDCQI_STATUS_UNKNOWN,White; end
	if (not LootCount_DropCount_DB.Quest) then return LCDCQI_STATUS_UNKNOWN,White; end

	-- Check running quests
	if (LootCount_DropCount_Character.Quests) then
		if (LootCount_DropCount_Character.Quests[checkquest]) then
			return LCDCQI_STATUS_STARTED,Yellow;	-- I have it
		end
	end
	-- Maybe it's done
	if (LootCount_DropCount_Character.DoneQuest) then
		if (LootCount_DropCount_Character.DoneQuest[checkquest]) then
			return LCDCQI_STATUS_DONE,Dark;		-- I've done it
		end
	end

	return LCDCQI_STATUS_NOTSTARTED,Green;
end


function LootCount_DropCount_SetQuestTT(faction,npc,parent)
	if (not LootCount_DropCount_DB.Quest) then return; end
	if (not LootCount_DropCount_DB.Quest[faction]) then return; end
	if (not LootCount_DropCount_DB.Quest[faction][npc]) then return; end
	if (not LootCount_DropCount_DB.Quest[faction][npc].Quests) then return; end
	LootCount_DropCount_TT:ClearLines();
	if (not parent) then parent=UIParent; end
	LootCount_DropCount_TT:SetOwner(parent,"ANCHOR_CURSOR");
	LootCount_DropCount_TT:SetText(npc);
	for _,quest in pairs(LootCount_DropCount_DB.Quest[faction][npc].Quests) do
		local _,colour=LootCount_DropCount_GetQuestStatus(quest);
		LootCount_DropCount_TT:AddLine("  "..colour..quest,1,1,1);
	end
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_SetBookTT(book,parent)
	if (not LootCount_DropCount_DB.Book) then return; end
	if (not LootCount_DropCount_DB.Book[book]) then return; end
	LootCount_DropCount_TT:ClearLines();
	if (not parent) then parent=UIParent; end
	LootCount_DropCount_TT:SetOwner(parent,"ANCHOR_CURSOR");
	LootCount_DropCount_TT:SetText("Book");
	LootCount_DropCount_TT:AddLine(book,1,1,1);
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_SetQuestlist(npc,parent)
	if (not LootCount_DropCount_DB.Quest) then return; end
	for faction,fTable in pairs(LootCount_DropCount_DB.Quest) do
		if (fTable[npc]) then
			LootCount_DropCount_TT:ClearLines();
			if (not parent) then parent=UIParent; end
			LootCount_DropCount_TT:SetOwner(parent,"ANCHOR_CURSOR");
			LootCount_DropCount_TT:SetText(npc);

			local line=1;
			while (fTable[npc].Quests[line]) do
				local _,colour=LootCount_DropCount_GetQuestStatus(fTable[npc].Quests[line]);
				LootCount_DropCount_TT:AddDoubleLine("  "..colour..fTable[npc].Quests[line],"",1,1,1,1,1,1);
				line=line+1;
			end
		end
	end
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_SetVendorlist(unit,parent)
	local breakit=nil;
	if (LootCount_DropCount_DB.Quest) then
		for faction,fTable in pairs(LootCount_DropCount_DB.Quest) do
			for npc,nTable in pairs(fTable) do
				if (npc==unit) then
					breakit=true;
					LootCount_DropCount_SetQuestlist(unit,parent);
				end
				if (breakit) then break; end
			end
			if (breakit) then break; end
		end
	end
	if (not LootCount_DropCount_DB.Vendor) then return; end
	if (not LootCount_DropCount_DB.Vendor[unit]) then return; end
	LootCount_DropCount_TT:ClearLines();
	if (not parent) then parent=UIParent; end
	LootCount_DropCount_TT:SetOwner(parent,"ANCHOR_CURSOR");
	local line=unit;
	if (LootCount_DropCount_DB.Vendor[unit].Repair) then line=line..LCDC_C_GREEN.." (Repair)"; end
	LootCount_DropCount_TT:SetText(line);

	local list={};
	line=1;
	local missingitems=0;
	local itemsinlist=nil;
	local item,iTable;
	for item,iTable in pairs(LootCount_DropCount_DB.Vendor[unit].Items) do
		local itemname,_,rarity=GetItemInfo(item);
		if (not itemname or not rarity) then
			LootCount_DropCount_StartCache(item);
			missingitems=missingitems+1;
		else
			local _,_,_,colour=GetItemQualityColor(rarity);
			list[line]={ Ltext=colour..itemname.."|r ", Rtext="" };
			if (iTable.Count>=0) then list[line].Ltext=LCDC_C_RED.."* |r"..list[line].Ltext;
			elseif (iTable.Count==LCDC_UNKNOWNCOUNT) then list[line].Ltext=LCDC_C_YELLOW.."* |r"..list[line].Ltext;
			end
			line=line+1;
			itemsinlist=true;
		end
	end
	if (missingitems>0) then
		LootCount_DropCount_TT:AddDoubleLine("Missing "..missingitems.." items.","",1,0,0,0,0,0);
		LootCount_DropCount_TT:AddDoubleLine("Loading...","",1,0,0,0,0,0);
		LootCount_DropCount_TT:Show();
		LootCount_DropCount_TT.Loading=true;
		return;
	end
	LootCount_DropCount_TT.Loading=nil;
	if (not itemsinlist) then LootCount_DropCount_TT:Hide(); return; end

	list=LootCount_DropCount_SortNames(list);
	line=1;
	while(list[line]) do
		LootCount_DropCount_TT:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
		line=line+1;
	end
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_SetLootlist(unit)
	if (not LootCount_DropCount_DB.Count[unit]) then
		LootCount_DropCount_SetVendorlist(unit);
		return;
	end

	LootCount_DropCount_TT:ClearLines();
	LootCount_DropCount_TT:SetOwner(UIParent,"ANCHOR_CURSOR");
	LootCount_DropCount_TT:SetText(unit);
	local text="";
	if (LootCount_DropCount_DB.Count[unit].Skinning) then text="Profession-loot: "..LootCount_DropCount_DB.Count[unit].Skinning.." times"; end
	LootCount_DropCount_TT:AddDoubleLine(LootCount_DropCount_DB.Count[unit].Kill.." kills",text,.4,.4,1,1,0,1);

	local list={};
	local line=1;
	local missingitems=0;
	local itemsinlist=nil;
	for item,iTable in pairs(LootCount_DropCount_DB.Item) do
		if (iTable.Name) then
			for mob,mTable in pairs(iTable.Name) do
				if (mob==unit) then
					local itemname,_,rarity=GetItemInfo(item);
					if (not itemname or not rarity) then
						LootCount_DropCount_StartCache(item);
						missingitems=missingitems+1;
					else
						local _,_,_,colour=GetItemQualityColor(rarity);
						local thisratio,_,thissafe=LootCount_DropCount_GetRatio(item,mob);
						list[line]={ Ltext=colour.."["..itemname.."]|r: ", ratio=thisratio, NoSafe=thissafe };
						line=line+1;
						itemsinlist=true;
					end
				end
			end
		end
		if (iTable.Skinning) then
			for mob,mTable in pairs(iTable.Skinning) do
				if (mob==unit) then
					local itemname,_,rarity=GetItemInfo(item);
					if (not itemname or not rarity) then
						LootCount_DropCount_StartCache(item);
						missingitems=missingitems+1;
					else
						local _,_,_,colour=GetItemQualityColor(rarity);
						local _,thisratio,thissafe=LootCount_DropCount_GetRatio(item,mob);
						list[line]={ Ltext=colour.."["..itemname.."]|r: ", ratio=thisratio, NoSafe=thissafe };
						if (LootCount_DropCount_DB.Item[item].Skinning) then
							list[line].Ltext="|cFFFF00FF*|r "..list[line].Ltext;	-- AARRGGBB
						end
						line=line+1;
						itemsinlist=true;
					end
				end
			end
		end
	end
	if (missingitems>0) then
		LootCount_DropCount_TT:AddDoubleLine("Missing "..missingitems.." items.","",1,0,0,0,0,0);
		LootCount_DropCount_TT:AddDoubleLine("Loading...","",1,0,0,0,0,0);
		LootCount_DropCount_TT:Show();
		LootCount_DropCount_TT.Loading=true;
		return;
	end
	LootCount_DropCount_TT.Loading=nil;
	if (not itemsinlist) then LootCount_DropCount_TT:Hide(); return; end

	list=LootCount_DropCount_SortRatio(list);
	line=1;
	while(list[line]) do
		LootCount_DropCount_TT:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
		line=line+1;
	end
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_StartCache(item)
	LootCount_DropCount_CacheRetries=0;
	if (not LootCount_DropCount_UnknownItems) then LootCount_DropCount_UnknownItems={}; LootCount_DropCount_EnumerateTimer=.5; end
	LootCount_DropCount_UnknownItems[item]=true;
end


-- A blind update will queue a request at the server without any
-- book-keeping at this side.
function LootCount_DropCount_CacheItem(item,blind)
	if (type(item)=="number") then item="item:"..item;
	elseif (type(item)~="string") then return true; end
	local name=GetItemInfo(item);
	if (not LootCount_DropCount_CF:IsVisible()) then
		if (not name) then
			LootCount_DropCount_CF:SetOwner(UIParent); LootCount_DropCount_CF:SetHyperlink(item); LootCount_DropCount_CF:Hide();
			if (not blind) then
				LootCount_DropCount_CacheRetries=LootCount_DropCount_CacheRetries+1;
			end
			return false;
		else
			if (not blind) then LootCount_DropCount_CacheRetries=0; end
		end
	end
	return true;
end


function LootCount_DropCount_ToggleMenu(button)
	LootCount_DropCount_This=button;
	ToggleDropDownMenu(nil,nil,LootCount_DropCount_MenuOptions,button,0,0);
end


function LootCount_DropCount_BuildMenu(button)
	LootCount_DropCount_This=button;
	UIDropDownMenu_Initialize(LootCount_DropCount_MenuOptions,LootCount_DropCount_MenuLoad,"MENU");
end


function LootCount_DropCount_MenuLoad()
	local info={};
	info.textR=1; info.textG=1; info.textB=1;
	if (LCT_QUEST and LootCount_DropCount_This.User and LootCount_DropCount_This.User.itemID) then
		local _,_,_,_,_,itemtype=GetItemInfo(LootCount_DropCount_This.User.itemID);
		if (itemtype and itemtype==LCT_QUEST) then
			info.isTitle=1; info.checked=nil; info.func=nil; info.text="Setting for this item:"; UIDropDownMenu_AddButton(info,1); info.isTitle=nil; info.disabled=nil;
			local _,thisid=DuckLib:GetID(LootCount_DropCount_This.User.itemID);
			if (not LootCount_DropCount_NoQuest[thisid]) then info.checked=true; end
			info.text="Only drops when on a quest"; info.value=thisid; info.func=LootCount_DropCount_Menu_ToggleQuestItem; UIDropDownMenu_AddButton(info,1);

			info.isTitle=1; info.checked=nil; info.func=nil; info.text=" "; UIDropDownMenu_AddButton(info,1);
		end
	end

	info.isTitle=1; info.checked=nil; info.func=nil; info.text="Global settings for "..LOOTCOUNT_DROPCOUNT_VERSIONTEXT..":"; UIDropDownMenu_AddButton(info,1); info.isTitle=nil; info.disabled=nil;

	info.func=LootCount_DropCount_Menu_ToggleZone;
	info.text="Show zone"; info.checked=LootCount_DropCount_Character.ShowZone; UIDropDownMenu_AddButton(info,1);
	info.func=LootCount_DropCount_Menu_ToggleZoneMobs;
	info.text="Only list mobs from current zone"; info.checked=LootCount_DropCount_Character.ShowZoneMobs; UIDropDownMenu_AddButton(info,1);
	info.func=LootCount_DropCount_Menu_SetGoal;
	info.text="Set goal"; info.checked=nil; UIDropDownMenu_AddButton(info,1);

	info.func=LootCount_DropCount_Menu_ToggleChannel;
	info.text="Guild data-sharing"; info.value="GUILD"; info.checked=LootCount_DropCount_DB[info.value]; UIDropDownMenu_AddButton(info,1);
	info.text="Group/raid data-sharing"; info.value="RAID"; info.checked=LootCount_DropCount_DB[info.value]; UIDropDownMenu_AddButton(info,1);
	info.checked=nil;
end


function LootCount_DropCount_Menu_SetGoal()
	LootCount_SetGoalPopup(LootCount_DropCount_This);
end


function LootCount_DropCount_Menu_ToggleZoneMobs()
	if (LootCount_DropCount_Character.ShowZoneMobs) then LootCount_DropCount_Character.ShowZoneMobs=false;
	else LootCount_DropCount_Character.ShowZoneMobs=true; end
end


function LootCount_DropCount_Menu_ToggleZone()
	if (LootCount_DropCount_Character.ShowZone) then LootCount_DropCount_Character.ShowZone=false;
	else LootCount_DropCount_Character.ShowZone=true; end
end


function LootCount_DropCount_Menu_ToggleQuestItem()
	if (LootCount_DropCount_NoQuest[this.value]) then LootCount_DropCount_NoQuest[this.value]=nil; return; end
	LootCount_DropCount_NoQuest[this.value]=true;
end


function LootCount_DropCount_Menu_ToggleChannel()
	if (LootCount_DropCount_DB[this.value]) then LootCount_DropCount_DB[this.value]=nil; else LootCount_DropCount_DB[this.value]=true; end
	if (this.value=="GUILD") then LootCount_DropCount_DB.RAID=nil; end
	if (this.value=="RAID") then LootCount_DropCount_DB.GUILD=nil; end
end


function LootCount_DropCount_Menu_SetGoal()
	if (not LootCount_DropCount_Registered) then return; end
	local user=LootCountAPI.User(LootCount_DropCount_This);
	LootCount_SetGoalPopup(LootCount_DropCount_This);
end


function LootCount_DropCount_CleanDB()
	local text=LCDC_C_BASIC.."DropCount:|r ";

	if (LootCount_DropCount_Debug) then
		collectgarbage("collect");
		UpdateAddOnMemoryUsage();
		local usage=GetAddOnMemoryUsage("LootCount_DropCount");
		usage=string.format("%.0fKB",usage);
		text="DropCount memory usage: "..usage.." -> ";
	end

	local nowitems,nowmobs;
	if (LootCount_DropCount_DB.Book) then
		nowitems=0;
		for vendor,vTable in pairs(LootCount_DropCount_DB.Book) do nowitems=nowitems+1; end
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_GREEN..nowitems.."|r known books");
	end
	if (LootCount_DropCount_DB.Vendor) then
		nowmobs=0;
		for vendor,vTable in pairs(LootCount_DropCount_DB.Vendor) do nowmobs=nowmobs+1; end
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_GREEN..nowmobs.."|r known vendors");
	end
	if (LootCount_DropCount_DB.Quest) then
		nowmobs=0;
		for _,fTable in pairs(LootCount_DropCount_DB.Quest) do
			for _,nTable in pairs(fTable) do
				for _,qTable in pairs(nTable.Quests) do
					nowmobs=nowmobs+1;
				end
			end
		end
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_GREEN..nowmobs.."|r known quests");
	end
	nowitems=0;
	if (LootCount_DropCount_DB.Item) then
		for item,iTable in pairs(LootCount_DropCount_DB.Item) do nowitems=nowitems+1; end
		text=text..LCDC_C_GREEN..nowitems.."|r items";
	end
	if (not LootCount_DropCount_StartItems) then LootCount_DropCount_StartItems=nowitems; end
	nowmobs=0;
	if (LootCount_DropCount_DB.Count) then
		for creature,cTable in pairs(LootCount_DropCount_DB.Count) do nowmobs=nowmobs+1; end
		text=text.." -> "..LCDC_C_GREEN..nowmobs.."|r creatures";
	end
	if (not LootCount_DropCount_StartMobs) then LootCount_DropCount_StartMobs=nowmobs; end
	DuckLib:Chat(text);

	if (nowitems-LootCount_DropCount_StartItems>0 or nowmobs-LootCount_DropCount_StartMobs>0) then
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r New this session: "..LCDC_C_GREEN..nowitems-LootCount_DropCount_StartItems.."|r items, "..LCDC_C_GREEN..nowmobs-LootCount_DropCount_StartMobs.."|r mobs");
	end
	DuckLib:Chat(LCDC_C_BASIC.."Type "..LCDC_C_GREEN..SLASH_DROPCOUNT2..LCDC_C_BASIC.." to view options");
end


function LCDC_ConvertBookFormat()
	if (not LootCount_DropCount_DB.Book) then return; end
	local book,bTable;
	for book,bTable in pairs(LootCount_DropCount_DB.Book) do
		if (LootCount_DropCount_DB.Book[book].Zone) then
			LootCount_DropCount_DB.Book[book][1]= {};
			LootCount_DropCount_DB.Book[book][1].X=LootCount_DropCount_DB.Book[book].X;
			LootCount_DropCount_DB.Book[book][1].Y=LootCount_DropCount_DB.Book[book].Y;
			LootCount_DropCount_DB.Book[book][1].Zone=LootCount_DropCount_DB.Book[book].Zone;
			LootCount_DropCount_DB.Book[book].X=nil;
			LootCount_DropCount_DB.Book[book].Y=nil;
			LootCount_DropCount_DB.Book[book].Zone=nil;
		end
	end
end


function LCDC_SaveBook(BookName)
	if (not LootCount_DropCount_DB.Book) then LootCount_DropCount_DB.Book={}; end

	local bX,bY=LCDC_GetPLayerPosition();
	local bZone=LootCount_DropCount_GetZone();
	if (not bZone) then bZone="Unknown"; end
	local i=1;
	if (not LootCount_DropCount_DB.Book[BookName]) then
		DuckLib:Chat(LCDC_C_BASIC.."Location of new book "..LCDC_C_GREEN.."\""..BookName.."\""..LCDC_C_BASIC.." saved");
		LootCount_DropCount_DB.Book[BookName]={};
	else
		local found=nil;
		while (not found and LootCount_DropCount_DB.Book[BookName][i]) do
			if (LootCount_DropCount_DB.Book[BookName][i].Zone==bZone) then		-- Found in same zone
				if (LootCount_DropCount_DB.Book[BookName][i].X<=(bX+1)) then
					if (LootCount_DropCount_DB.Book[BookName][i].X>=(bX-1)) then
						if (LootCount_DropCount_DB.Book[BookName][i].Y<=(bY+1)) then
							if (LootCount_DropCount_DB.Book[BookName][i].Y>=(bY-1)) then
								found=true;
								i=i-1;
							end
						end
					end
				end
			end
			i=i+1;
		end
	end
	LootCount_DropCount_DB.Book[BookName][i]={
		Zone=bZone,
		X=bX,
		Y=bY,
	};
	LootCount_DropCount_MakeIconsBook();
	DuckLib:Chat(LCDC_C_BASIC..BookName..LCDC_C_GREEN.." saved. "..LCDC_C_BASIC..i.." volumes known.");
end


function LootCount_DropCount_OnUpdate(elapsed)
	LCDC_LastUpdate=LCDC_LastUpdate+elapsed;
	if (LCDC_LastUpdate<(1/20)) then return; end
	LCDC_LastUpdate=0;

	if (not DuckLib) then return; end				-- Library is missing

	if (not LCT_QUEST and (not LootCount_DropCount_UnknownItems or not LootCount_DropCount_UnknownItems["item:31812"])) then
		_,_,_,_,_,LCT_QUEST=GetItemInfo("item:31812");
		if (not LCT_QUEST) then
			LootCount_DropCount_StartCache("item:31812")
		else
			LootCount_DropCount_ConvertDB();
		end
	end

	if (LootCount_DropCount_UnknownItems) then
		LootCount_DropCount_EnumerateTimer=LootCount_DropCount_EnumerateTimer-elapsed;
		if (LootCount_DropCount_EnumerateTimer<=0) then
			LootCount_DropCount_EnumerateTimer=LCDC_ENUMERATE;
			if (not LootCount_DropCount_RunningItem) then
				local counter=0;
				for unknown,value in pairs(LootCount_DropCount_UnknownItems) do
					if (not LootCount_DropCount_RunningItem) then
						LootCount_DropCount_RunningItem=unknown;
					end
					LootCount_DropCount_CacheItem(unknown,true);	-- Blind (pre) update
					counter=counter+1;
					if (counter>=5) then break; end
				end
--DuckLib:Chat(counter.." - "..LootCount_DropCount_RunningItem,1);
				if (not LootCount_DropCount_RunningItem) then LootCount_DropCount_UnknownItems=nil; return; end
			end
			local itemname=GetItemInfo(LootCount_DropCount_RunningItem);
			if (not itemname) then
				LootCount_DropCount_CacheItem(LootCount_DropCount_RunningItem);
			else
				LootCount_DropCount_UnknownItems[LootCount_DropCount_RunningItem]=nil;
				if (LootCount_DropCount_QueryItem==LootCount_DropCount_RunningItem) then
					LootCount_DropCount_ItemVendors();
				end
				LootCount_DropCount_RunningItem=nil;
				LootCount_DropCount_CacheRetries=0;
				if (not LootCount_DropCount_DB.Converted) then
					-- Counting items
					local val,vTable,i;
					i=0;
					for val,vTable in pairs(LootCount_DropCount_UnknownItems) do i=i+1; end
					if (i<6 or (i>5 and (floor(i/25)*25)==i)) then
						DuckLib:Chat("Converting DropCount database: "..i.." items left...",0,1,1);
					end
				end
			end
		end
		if (LootCount_DropCount_CacheRetries>=LCSV_MAXRETRIES) then
			if (LootCount_DropCount_QueryItem==LootCount_DropCount_RunningItem) then
				DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_RED.."Could not retrieve information for this item from the server.");
			end
			DuckLib:Chat("Not retrievable: "..LootCount_DropCount_RunningItem,1);
			if (LootCount_DropCount_DB.Item and
				LootCount_DropCount_DB.Item[LootCount_DropCount_RunningItem] and
				LootCount_DropCount_DB.Item[LootCount_DropCount_RunningItem].Item) then
				DuckLib:Chat("\""..LootCount_DropCount_RunningItem.."\" seem to be \""..LootCount_DropCount_DB.Item[LootCount_DropCount_RunningItem].Item.."\"",1);
				DuckLib:Chat("This can happen if it has not been seen on the server since last server restart.",1);
			end
			LCDC_MaintainerPOS=-1;
			LootCount_DropCount_UnknownItems=nil;			-- Too many tries, so abort
			LootCount_DropCount_RunningItem=nil;
		end
	elseif (LCDC_CachedConvertItems>0) then
		LootCount_DropCount_ConvertDB();
	end

	if (LootCount_DropCount_VendorDelay>=0) then
		LootCount_DropCount_VendorDelay=LootCount_DropCount_VendorDelay-elapsed;
		if (LootCount_DropCount_VendorDelay<0) then
			if (not LootCount_DropCount_ReadMerchant(LootCount_DropCount_OpenMerchant)) then
				LootCount_DropCount_VendorDelay=.5;
			end
		end
	end

	if (IsAltKeyDown() and UnitExists("mouseover")) then
		if (not LootCount_DropCount_TT:IsVisible()) then LootCount_DropCount_SetLootlist(UnitName("mouseover"));
		elseif (LootCount_DropCount_TT.Loading) then
			LootCount_DropCount_TT:Hide();
			LootCount_DropCount_SetLootlist(UnitName("mouseover"));
		end
	elseif (not LCDC_VendorFlag_Info) then
		if (LootCount_DropCount_TT:IsVisible()) then LootCount_DropCount_TT:Hide(); end
	end


	if (LootCount_DropCount_QueueSize>0) then
		local now=time();
		if (LootCount_DropCount_QueueSize>0 and now-LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize].Time>LCDC_QUEUETIME) then
			LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize]=nil;
			LootCount_DropCount_QueueSize=LootCount_DropCount_QueueSize-1;
			if (LootCount_DropCount_Registered) then LootCountAPI.Force(LOOTCOUNT_DROPCOUNT); end
		end
	end

	if (GameTooltip and GameTooltip:IsVisible()) then
		local _,ThisItem=GameTooltip:GetItem();
		if (ThisItem) then
			ThisItem=DuckLib:GetID(ThisItem);
			if (ThisItem) then
				if (IsControlKeyDown()) then LootCount_DropCount_TooltipVendor(ThisItem);
				elseif (IsAltKeyDown()) then LootCount_DropCount_Tooltip(ThisItem); end
			end
		end
	end

	if (ItemTextFrame and ItemTextFrame:IsVisible()) then
		if (not LCDC_ItemTextFrameHandled) then
			if (ItemTextGetCreator()) then							-- A player created text
				LCDC_ItemTextFrameHandled=true;						-- Do nothing more
			elseif (ItemTextTitleText) then							-- There's a header
				local material=ItemTextGetMaterial();
				if (not material) then material="Parchment"; end
				if (material=="Parchment") then						-- It's patchment
					if (ItemTextNextPageButton:IsVisible()) then	-- Multi-page
						local theItem=ItemTextTitleText:GetText();
						if (theItem) then							-- There's something in the header
							if (not GetItemInfo(theItem)) then		-- It's not an item name
								LCDC_SaveBook(theItem);
								LCDC_ItemTextFrameHandled=true;		-- Do nothing more
							end
						end
					end
				end
			end
		end
	elseif (LCDC_ItemTextFrameHandled) then
		LCDC_ItemTextFrameHandled=nil;
	end

	if (LCDC_StartupDelay and LCDC_StartupDelay>0) then
		LCDC_StartupDelay=LCDC_StartupDelay-elapsed;
		if (LCDC_StartupDelay<=0) then
			LCDC_StartupDelay=nil;
			Astrolabe:CalculateMinimapIconPositions();
			LootCount_DropCount_CleanDB();
			LootCount_DropCount_Maintainer(true);
--			if (not LootCount_DropCount_Character.Quests) then		-- init current q-list
				LCDC_RescanQuests=LCDC_RESCANQUESTS;
--			end
		end
	end

	if (LCDC_RescanQuests) then
		LCDC_RescanQuests=LCDC_RescanQuests-elapsed;
		if (LCDC_RescanQuests<0) then
			LCDC_RescanQuests=nil;
			LCDC_ScanQuests();
			LootCount_DropCount_MakeIconsQuest();
		end
	end

	if (LCDC_MaintainerPOS>0) then LootCount_DropCount_Maintainer(); end

	if (QuestsDone.Waiter>=0) then
		QuestsDone.Waiter=QuestsDone.Waiter-elapsed;
		if (QuestsDone.Waiter<0 and LootCount_DropCount_DB.QuestQuery) then
			local count=nil;
			for qIndex,_ in pairs(LootCount_DropCount_DB.QuestQuery) do
				count=true;
--DuckLib:Chat("Checking: "..qIndex,1);
				local qName=LootCount_DropCount_GetQuestName("quest:"..qIndex);
				if (qName) then
					if (LootCount_DropCount_Character.DoneQuest[qName]) then
						if (type(LootCount_DropCount_Character.DoneQuest[qName])~="table") then
							if (LootCount_DropCount_Character.DoneQuest[qName]~=true) then
								local num=LootCount_DropCount_Character.DoneQuest[qName];
								LootCount_DropCount_Character.DoneQuest[qName]=nil;
								LootCount_DropCount_Character.DoneQuest[qName]={};
								LootCount_DropCount_Character.DoneQuest[qName][num]=true;
								LootCount_DropCount_Character.DoneQuest[qName][qIndex]=true;
							else
								LootCount_DropCount_Character.DoneQuest[qName]=qIndex;
							end
						else
							LootCount_DropCount_Character.DoneQuest[qName][qIndex]=true;
						end
					else
						LootCount_DropCount_Character.DoneQuest[qName]=qIndex;
					end
					LootCount_DropCount_DB.QuestQuery[qIndex]=nil;
					LCDC_QuestConvertItems=LCDC_QuestConvertItems-1;
--DuckLib:Chat(qIndex.." = "..qName.." ("..LCDC_QuestConvertItems..")",1);
				end
				QuestsDone.Waiter=(1/3);
				break;
			end
			if (not count) then
				LootCount_DropCount_DB.QuestQuery=nil;
				LCDC_RescanQuests=3;
			end
		end
	end


	if (LootCount_DropCount_Registered) then
-- Old fight-code removed
	elseif (LootCountAPI and LootCountAPI.Register and LootCount_Loaded) then
		-- Old LootCount dependency load-code
		local info = {	Name=LOOTCOUNT_DROPCOUNT,
						MenuText="LootCount Drop-rate",
						Update=LootCount_DropCount_UpdateButton,
						Drop=LootCount_DropCount_DropItem,
						Clicker=LootCount_DropCount_IconClicked,
						Texture="Interface\\Icons\\INV_Misc_QuestionMark",
						Tooltip=LootCount_DropCount_Tooltip,
						ColorBottom={ r=1, g=1, b=0 },
						ColorTop={ r=.7, g=.7, b=0 },
					};

		LootCount_DropCount_DB.RAID=nil;
		if (IsInGuild()) then LootCount_DropCount_DB.GUILD=true; else LootCount_DropCount_DB.GUILD=nil; end
		LootCountAPI.Register(info);
		LootCount_DropCount_Registered=true;
		DuckLib:Chat(LCDC_C_BASIC.."DropCount:|r "..LCDC_C_GREEN.."LootCount detected. DropCount is available from the LootCount menu.");
	end
end


function LootCount_DropCount_MergeDatabase()
	if (not LootCount_DropCount_MergeData) then return; end
	if (not LootCount_DropCount_DB.MergedData) then LootCount_DropCount_DB.MergedData=0; end
	if (LootCount_DropCount_MergeData.Version==LootCount_DropCount_DB.MergedData) then return; end

	-- Vendors
	local vendor={ New={ Unknown=0, }, Updated={ Unknown=0, }, };
	if (LootCount_DropCount_MergeData.Vendor) then
		if (not LootCount_DropCount_DB.Vendor) then LootCount_DropCount_DB.Vendor={}; end
		for vend,vTable in pairs(LootCount_DropCount_MergeData.Vendor) do
			local faction=vTable.Faction;
			if (not faction) then
				if (LootCount_DropCount_DB.Vendor[vend] and
					LootCount_DropCount_DB.Vendor[vend].Faction) then
					faction=LootCount_DropCount_DB.Vendor[vend].Faction;	-- Set old faction
				else
					faction="Unknown";										-- Unknown faction
				end
			end
			if (not vendor.New[faction]) then vendor.New[faction]=0; end
			if (not vendor.Updated[faction]) then vendor.Updated[faction]=0; end
			if (not LootCount_DropCount_DB.Vendor[vend]) then
				LootCount_DropCount_DB.Vendor[vend]=DuckLib:CopyTable(vTable);
				vendor.New[faction]=vendor.New[faction]+1;
			else
				local updated=nil;
				local tv=LootCount_DropCount_DB.Vendor[vend];
				if (vTable.X>0 or vTable.Y>0) then
					if (math.floor(tv.X)~=math.floor(vTable.X) or
						math.floor(tv.Y)~=math.floor(vTable.Y) or
						tv.Zone~=vTable.Zone) then
						updated=true;
					end
					tv.X=vTable.X; tv.Y=vTable.Y; tv.Zone=vTable.Zone;
					if (vTable.Faction) then tv.Faction=vTable.Faction; end
				end
				if (vTable.Items) then
					if (not tv.Items) then
						tv.Items=DuckLib:CopyTable(vTable.Items);
						updated=true;
					else
						for item,iTable in pairs(vTable.Items) do
							if (not tv.Items[item]) then
								tv.Items[item]=DuckLib:CopyTable(iTable);
								updated=true;
							else
								if (iTable.Count~=-2 and tv.Items[item].Count==-2) then
									tv.Items[item].Count=iTable.Count;
									updated=true;
								end
							end
						end
					end
				end
				if (updated) then vendor.Updated[faction]=vendor.Updated[faction]+1; end
			end
		end
	end

	-- Books
	local book={ New=0, Updated=0, };
	if (LootCount_DropCount_MergeData.Book) then
		if (not LootCount_DropCount_DB.Book) then LootCount_DropCount_DB.Book={}; end
		-- Traverse hardcoded
		for title,bTable in pairs(LootCount_DropCount_MergeData.Book) do
			if (not LootCount_DropCount_DB.Book[title]) then
				-- Don't have it. Take all
				LootCount_DropCount_DB.Book[title]=DuckLib:CopyTable(bTable);
				book.New=book.New+1;
			else
				-- Have it. Check volumes
				local tb=LootCount_DropCount_DB.Book[title];
				for volume,vTable in pairs(bTable) do		-- Traverse hardcoded volumes
					local saved=nil;
					for index,iTable in pairs(tb) do		-- Traverse my own volumes
						if (vTable.Zone==iTable.Zone) then
							if (vTable.X <= iTable.X+1) then
								if (vTable.X >= iTable.X-1) then
									if (vTable.Y <= iTable.Y+1) then
										if (vTable.Y >= iTable.Y-1) then
											iTable.X=vTable.X;
											iTable.Y=vTable.Y;
											saved=true;
											break;
										end
									end
								end
							end
						end
					end
					-- Match not found
					if (not saved) then
						local index=1;
						while (tb[index]) do index=index+1; end		-- Find end
						tb[index]=DuckLib:CopyTable(vTable);		-- Insert new volume
						book.Updated=book.Updated+1;
					end
				end
			end
		end
	end


	-- Quests
	local questNPC={ New={ Unknown=0, }, Updated={ Unknown=0, }, };
	if (LootCount_DropCount_MergeData.Quest) then
		if (not LootCount_DropCount_DB.Quest) then LootCount_DropCount_DB.Quest={}; end
		-- Traverse hardcoded
		for faction,fTable in pairs(LootCount_DropCount_MergeData.Quest) do
			if (not LootCount_DropCount_DB.Quest[faction]) then
				-- Don't have this faction, so take all.
				LootCount_DropCount_DB.Quest[faction]=DuckLib:CopyTable(fTable);
			else
				questNPC.New[faction]=0;
				questNPC.Updated[faction]=0;
				-- Traverse hardcoded npcs in this faction
				for npc,nTable in pairs(fTable) do
					if (not LootCount_DropCount_DB.Quest[faction][npc]) then
						-- Don't have it, so take all
						LootCount_DropCount_DB.Quest[faction][npc]=DuckLib:CopyTable(nTable);
						questNPC.New[faction]=questNPC.New[faction]+1;
					else
						local updated=nil;
						-- Have it, so update location and merge quests
						local tn=LootCount_DropCount_DB.Quest[faction][npc];
						tn.X=nTable.X; tn.Y=nTable.Y; tn.Zone=nTable.Zone;
						if (not tn.Quests) then tn.Quests={}; end
						local fromIndex=1;
						while (nTable.Quests[fromIndex]) do
							local toIndex=1;
							while (tn.Quests[toIndex]) do
								if (tn.Quests[toIndex]==nTable.Quests[fromIndex]) then
									toIndex=1000;		-- It is already here
								end
								toIndex=toIndex+1;
							end
							-- Check if it was already there or not
							if (toIndex<1000) then
								tn.Quests[toIndex]=nTable.Quests[fromIndex];
								updated=true;
							end
							fromIndex=fromIndex+1;
						end
						if (updated) then questNPC.Updated[faction]=questNPC.Updated[faction]+1; end
					end
				end
			end
		end
	end

	-- Merge drops
	local mobs={ New=0, Updated=0, };
	if (LootCount_DropCount_MergeData.Count and LootCount_DropCount_MergeData.Item) then
		if (not LootCount_DropCount_DB.Count) then LootCount_DropCount_DB.Count={}; end
		if (not LootCount_DropCount_DB.Item) then LootCount_DropCount_DB.Item={}; end
		for mob,mTable in pairs(LootCount_DropCount_MergeData.Count) do
			local newMob,updatedMob=LootCount_DropCount_MergeMOB(mob);
			mobs.New=mobs.New+newMob;
			mobs.Updated=mobs.Updated+updatedMob;
		end
	end

	-- Output result
	LootCount_DropCount_DB.MergedData=LootCount_DropCount_MergeData.Version;
	DuckLib:Chat(LOOTCOUNT_DROPCOUNT_VERSIONTEXT);
	DuckLib:Chat("Data merge summary:");
	-- mobs
	if (mobs.New>0) then DuckLib:Chat(mobs.New.." new mobs"); end
	if (mobs.Updated>0) then DuckLib:Chat(mobs.New.." updated mobs"); end
	-- Vendors
	for faction,fValue in pairs(vendor.New) do
		if (faction~="Unknown") then DuckLib:Chat("New vendors for "..faction..": "..fValue); end
	end
	if (vendor.New.Unknown>0) then DuckLib:Chat(vendor.New.Unknown.." new vendors with unknown faction"); end
	for faction,fValue in pairs(vendor.Updated) do
		if (faction~="Unknown") then DuckLib:Chat("Updated vendors for "..faction..": "..fValue); end
	end
	if (vendor.Updated.Unknown>0) then DuckLib:Chat(vendor.New.Unknown.." updated vendors with unknown faction"); end
	-- Quests
	for faction,fValue in pairs(questNPC.New) do
		if (faction~="Unknown") then DuckLib:Chat("New quest-givers for "..faction..": "..fValue); end
	end
	if (questNPC.New.Unknown>0) then DuckLib:Chat(vendor.New.Unknown.." new quest-givers with unknown faction"); end
	for faction,fValue in pairs(questNPC.Updated) do
		if (faction~="Unknown") then DuckLib:Chat("Updated quest-givers for "..faction..": "..fValue); end
	end
	if (questNPC.Updated.Unknown>0) then DuckLib:Chat(vendor.New.Unknown.." updated quest-givers with unknown faction"); end
	-- Books
	if (book.New>0) then DuckLib:Chat(book.New.." new books"); end
	if (book.Updated>0) then DuckLib:Chat(book.New.." updated books"); end
end


-- Check mobs and insert anything that is missing
-- IMPORTANT: Do not merge counts! Only insert missing!
--
-- o Check if new data has more kills/skinnings than me
--   If I don't have it at all, mine is set up as zero.
function LootCount_DropCount_MergeMOB(mob)
	local newMob,updatedMob=0,0;
	local kill,skinning=nil,nil;

	if (LootCount_DropCount_MergeData.Count[mob].Kill) then
		-- If it's not there, insert with no loot
		if (not LootCount_DropCount_DB.Count[mob]) then
			LootCount_DropCount_DB.Count[mob]=DuckLib:CopyTable(LootCount_DropCount_MergeData.Count[mob]);
			LootCount_DropCount_DB.Count[mob].Kill=0;
			LootCount_DropCount_DB.Count[mob].Skinning=0;
			newMob=1;
		end
		if (LootCount_DropCount_MergeData.Count[mob].Kill and
			LootCount_DropCount_DB.Count[mob].Kill) then
			if (LootCount_DropCount_MergeData.Count[mob].Kill>LootCount_DropCount_DB.Count[mob].Kill) then
				kill=LootCount_DropCount_MergeData.Count[mob].Kill;
				LootCount_DropCount_DB.Count[mob].Kill=kill;
				if (newMob==0) then updatedMob=1; end
			end
		end
		if (LootCount_DropCount_MergeData.Count[mob].Skinning and
			LootCount_DropCount_DB.Count[mob].Skinning) then
			if (LootCount_DropCount_MergeData.Count[mob].Skinning>LootCount_DropCount_DB.Count[mob].Skinning) then
				skinning=LootCount_DropCount_MergeData.Count[mob].Skinning;
				LootCount_DropCount_DB.Count[mob].Skinning=skinning;
				if (newMob==0) then updatedMob=1; end
			end
		end
	end

	-- Traverse hardcoded items
	for item,iTable in pairs(LootCount_DropCount_MergeData.Item) do
		-- Do normal kill/loot
		if (kill) then
			if (iTable.Name and iTable.Name[mob]) then	-- Exists in source
				if (not LootCount_DropCount_DB.Item[item]) then		-- Unknown item in target
					LootCount_DropCount_DB.Item[item]=DuckLib:CopyTable(iTable);	-- Copy item to target
					if (LootCount_DropCount_DB.Item[item].Name) then	-- There are mobs
						-- These moust be removed, as it may contain something
						-- that is should not be included. Never include unless
						-- the "Count"-table says so.
						DuckLib:ClearTable(LootCount_DropCount_DB.Item[item].Name);	-- Remove them
					end
				end
				if (not LootCount_DropCount_DB.Item[item].Name) then	-- Check for mob-list
					LootCount_DropCount_DB.Item[item].Name={};			-- Make it if not there
				end
				LootCount_DropCount_DB.Item[item].Name[mob]=iTable.Name[mob];	-- Insert this mob
			end
		end
		-- Do profession-loot
		if (skinning) then
			if (iTable.Skinning and iTable.Skinning[mob]) then	-- Exists in source
				if (not LootCount_DropCount_DB.Item[item]) then		-- Unknown item in target
					LootCount_DropCount_DB.Item[item]=DuckLib:CopyTable(iTable);	-- Copy item to target
					if (LootCount_DropCount_DB.Item[item].Skinning) then	-- There are mobs
						DuckLib:ClearTable(LootCount_DropCount_DB.Item[item].Skinning);	-- Remove them
					end
				end
				if (not LootCount_DropCount_DB.Item[item].Skinning) then	-- Check for mob-list
					LootCount_DropCount_DB.Item[item].Skinning={};			-- Make it if not there
				end
				LootCount_DropCount_DB.Item[item].Skinning[mob]=iTable.Skinning[mob];	-- Insert this mob
			end
		end
	end

	return newMob,updatedMob;
end

--[[
	Item
		["item:25450:0:0:0:0:0:0"] = {
			["Item"] = "Creeping Moss",
			["Time"] = 1270462376,
			["Name"] = {
				["Withered Bog Lord"] = 69,
				["Fungal Giant"] = 8,
				["Starving Bog Lord"] = 6,
				["Starving Fungal Giant"] = 5,
				["Bog Lord"] = 13,
				["Withered Giant"] = 116,
			},
			["Skinning"] = {
				["Withered Giant"] = 10,
				["Withered Bog Lord"] = 8,
			},
		},
	Count
		["Withered Bog Lord"] = {
			["Skinning"] = 29,
			["Kill"] = 94,
			["Zone"] = "Zangarmarsh - The Dead Mire",
		},
]]



function LCDC_Menu_AddHeader(text,icon)
	local info=UIDropDownMenu_CreateInfo();
	info.text=LCDC_C_LBLUE..text;
	info.icon=icon;
	UIDropDownMenu_AddButton(info,1);
end
function LCDC_Menu_AddChecker(text,value,func,icon)
	local info=UIDropDownMenu_CreateInfo();
	info.text=LCDC_C_BASIC..text;
	if (value) then info.text=info.text..LCDC_C_GREEN.."ON"; else info.text=info.text..LCDC_C_RED.."OFF"; end
	info.func=func;
	info.icon=icon;
	UIDropDownMenu_AddButton(info,1);
end

function LCDC_Minimap_Menu_Initialise()
	LCDC_Menu_AddHeader("Minimap");
	LCDC_Menu_AddChecker("Vendors: ",LootCount_DropCount_DB.VendorMinimap,LCDC_Settings_ToggleVendorsMinimap,"Interface\\GROUPFRAME\\UI-Group-MasterLooter");
	LCDC_Menu_AddChecker("Books: ",LootCount_DropCount_DB.BookMinimap,LCDC_Settings_ToggleBookMinimap,"Interface\\Spellbook\\Spellbook-Icon");
	LCDC_Menu_AddChecker("Quests: ",LootCount_DropCount_DB.QuestMinimap,LCDC_Settings_ToggleQuestMinimap,"Interface\\Spellbook\\Spellbook-Icon");

	LCDC_Menu_AddHeader(" ");
	LCDC_Menu_AddHeader("Worldmap");
	LCDC_Menu_AddChecker("Vendors: ",LootCount_DropCount_DB.VendorWorldmap,LCDC_Settings_ToggleVendorsWorldmap,"Interface\\GROUPFRAME\\UI-Group-MasterLooter");
	LCDC_Menu_AddChecker("Books: ",LootCount_DropCount_DB.BookWorldmap,LCDC_Settings_ToggleBookWorldmap,"Interface\\Spellbook\\Spellbook-Icon");
	LCDC_Menu_AddChecker("Quests: ",LootCount_DropCount_DB.QuestWorldmap,LCDC_Settings_ToggleQuestWorldmap,"Interface\\Spellbook\\Spellbook-Icon");
end


function LCDC_Settings_ToggleVendorsMinimap()
	if (LootCount_DropCount_DB.VendorMinimap) then LootCount_DropCount_DB.VendorMinimap=nil;
	else LootCount_DropCount_DB.VendorMinimap=true; end
	LootCount_DropCount_MakeIcons();
end
function LCDC_Settings_ToggleVendorsWorldmap()
	if (LootCount_DropCount_DB.VendorWorldmap) then LootCount_DropCount_DB.VendorWorldmap=nil;
	else LootCount_DropCount_DB.VendorWorldmap=true; end
end
function LCDC_Settings_ToggleBookMinimap()
	if (LootCount_DropCount_DB.BookMinimap) then LootCount_DropCount_DB.BookMinimap=nil;
	else LootCount_DropCount_DB.BookMinimap=true; end
	LootCount_DropCount_MakeIconsBook();
end
function LCDC_Settings_ToggleBookWorldmap()
	if (LootCount_DropCount_DB.BookWorldmap) then LootCount_DropCount_DB.BookWorldmap=nil;
	else LootCount_DropCount_DB.BookWorldmap=true; end
end
function LCDC_Settings_ToggleQuestMinimap()
	if (LootCount_DropCount_DB.QuestMinimap) then LootCount_DropCount_DB.QuestMinimap=nil;
	else LootCount_DropCount_DB.QuestMinimap=true; end
	LootCount_DropCount_MakeIconsQuest();
end
function LCDC_Settings_ToggleQuestWorldmap()
	if (LootCount_DropCount_DB.QuestWorldmap) then LootCount_DropCount_DB.QuestWorldmap=nil;
	else LootCount_DropCount_DB.QuestWorldmap=true; end
end


--[[    Minimap icon stuff    ]]
function LCDC_Minimap_OnEnter()
	GameTooltip:SetOwner(this,"ANCHOR_LEFT");
	GameTooltip:SetText("DropCount");
	GameTooltipTextLeft1:SetTextColor(0,1,0);
	GameTooltip:AddLine(LOOTCOUNT_DROPCOUNT_VERSIONTEXT);
	GameTooltip:AddLine(LCDC_C_BASIC.."<Left-click>|r for menu");
	GameTooltip:AddLine(LCDC_C_BASIC.."<Right-click>|r and drag to move");
	GameTooltip:AddLine(LCDC_C_BASIC.."<Shift-right-click>|r and drag for free-move");
	if (LootCount_DropCount_Registered) then
		GameTooltip:AddLine(LCDC_C_BASIC.."LootCount: "..LCDC_C_GREEN.."Present");
	end
	GameTooltip:Show();
end

function LCDC_Minimap_OnClick()
	ToggleDropDownMenu(1,nil,LCDC_Minimap_Menu,this:GetParent(),0,0);
end


-- Thanks to Yatlas and Gello for the initial code
function LCDC_Minimap_BeingDragged()
	-- Thanks to Gello for this code
	local xpos,ypos=GetCursorPosition();
	local xmin,ymin=Minimap:GetLeft(),Minimap:GetBottom();

	if (IsShiftKeyDown()) then
		LootCount_DropCount_DB.IconPosition=nil;
		xpos=(xpos/UIParent:GetScale()-xmin)-16;
		ypos=(ypos/UIParent:GetScale()-ymin)+16;
		LCDC_Minimap_SetIconAbsolute(xpos,ypos);
		return;
	end
	LootCount_DropCount_DB.IconX=nil;
	LootCount_DropCount_DB.IconY=nil;

	xpos=xmin-xpos/UIParent:GetScale()+70
	ypos=ypos/UIParent:GetScale()-ymin-70

	LCDC_Minimap_SetIconAngle(math.deg(math.atan2(ypos,xpos)));
end

function LCDC_Minimap_SetIconAngle(v)
	if (v<0) then v=v+360; end
	if (v>=360) then v=v-360; end

	LootCount_DropCount_DB.IconPosition=v;
	DropCount_MinimapIcon:SetPoint("TOPLEFT","Minimap","TOPLEFT",54-(78*cos(LootCount_DropCount_DB.IconPosition)),(78*sin(LootCount_DropCount_DB.IconPosition))-55);
	DropCount_MinimapIcon:Show();
end

function LCDC_Minimap_SetIconAbsolute(x,y)
	LootCount_DropCount_DB.IconX=x;
	LootCount_DropCount_DB.IconY=y;
	DropCount_MinimapIcon:SetPoint("TOPLEFT","Minimap","BOTTOMLEFT",x,y);
end
